try:
    from . import generic as g
except BaseException:
    import generic as g


class RasterTest(g.unittest.TestCase):

    def test_rasterize(self):
        p = g.get_mesh('2D/wrench.dxf')

        origin = p.bounds[0]
        pitch = p.extents.max() / 600
        resolution = g.np.ceil(p.extents / pitch).astype(int)

        # rasterize with filled
        filled = p.rasterize(origin=origin,
                             pitch=pitch,
                             resolution=resolution,
                             fill=True,
                             width=None)

        # rasterize just the outline
        outline = p.rasterize(origin=origin,
                              pitch=pitch,
                              resolution=resolution,
                              fill=False,
                              width=2.0)

        # rasterize both
        both = p.rasterize(origin=origin,
                           pitch=pitch,
                           resolution=resolution,
                           fill=True,
                           width=2.0)

        # rasterize with two-dimensional pitch
        pitch = p.extents / 600
        filled_2dpitch = p.rasterize(origin=origin,
                                     pitch=pitch,
                                     resolution=resolution,
                                     fill=True,
                                     width=None)

        # count the number of filled pixels
        fill_cnt = g.np.array(filled).sum()
        fill_2dpitch_cnt = g.np.array(filled_2dpitch).sum()
        both_cnt = g.np.array(both).sum()
        outl_cnt = g.np.array(outline).sum()

        # filled should have more than an outline
        assert fill_cnt > outl_cnt
        # filled+outline should have more than outline
        assert both_cnt > outl_cnt
        # filled+outline should have more than filled
        assert both_cnt > fill_cnt
        # A different pitch results in a different image
        assert fill_2dpitch_cnt != fill_cnt

    def test_nested(self):
        # make a test path with nested circles
        theta = g.np.linspace(0, g.np.pi * 2, 100)
        unit = g.np.column_stack((g.np.cos(theta), g.np.sin(theta)))
        radii = g.np.linspace(1.0, 10.0, 10)
        g.np.random.shuffle(radii)
        paths = []
        for R in radii:
            paths.append(g.trimesh.load_path(R * unit))
        path = g.trimesh.path.util.concatenate(paths)

        # split and extrude should both show 5 regions
        assert len(path.split()) == 5
        assert len(path.extrude(1.0)) == 5

        pitch = path.extents.max() / 1000
        origin = path.bounds[0] - pitch
        resolution = (g.np.ceil(
            path.extents / pitch) + 2).astype(int)

        # rasterize using the settings
        r = path.rasterize(
            pitch=pitch, origin=origin, resolution=resolution)
        # it's a boolean image so filled cells times
        # pitch area should be about the same as the area
        filled = g.np.array(r).sum() * pitch ** 2

        assert g.np.isclose(filled, path.area, rtol=0.01)


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class DepTest(g.unittest.TestCase):

    def test_deprecated(self):

        tests = [g.get_mesh('2D/wrench.dxf'),
                 g.trimesh.creation.box()]

        # todo : properly hash transform trees
        # so that copies of scenes hash the same
        # g.get_mesh('cycloidal.3DXML')]

        for m in tests:
            copy = m.copy()
            # the modern cool way of hashing
            assert hash(m) == hash(copy)
            assert m.__hash__() == copy.__hash__()
            assert m.identifier_hash == copy.identifier_hash

            # october 2023 deprecated ways of hashing
            # geometries
            assert m.md5() == copy.md5()
            assert m.crc() == copy.crc()
            assert m.hash() == copy.hash()
            assert m.identifier_md5 == copy.identifier_md5
            # trackedarray
            assert m.vertices.md5() == copy.vertices.md5()
            assert m.vertices.hash() == copy.vertices.hash()
            assert m.vertices.crc() == copy.vertices.crc()


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class DecompositionTest(g.unittest.TestCase):

    def test_convex_decomposition(self):
        mesh = g.get_mesh('quadknot.obj')

        engines = [('vhacd', g.trimesh.interfaces.vhacd.exists)]

        for engine, exists in engines:
            if not exists:
                g.log.warning(
                    'skipping convex decomposition engine %s', engine)
                continue

            g.log.info('Testing convex decomposition with engine %s', engine)
            meshes = mesh.convex_decomposition(engine=engine)
            self.assertTrue(len(meshes) > 1)
            for m in meshes:
                self.assertTrue(m.is_watertight)

            g.log.info('convex decomposition succeeded with %s', engine)


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g

from trimesh.viewer.trackball import Trackball


class TrackballTest(g.unittest.TestCase):

    def test_resize(self):
        trackball = Trackball(
            pose=g.np.eye(4),
            size=(640, 480),
            scale=1.0,
            target=g.np.array([0, 0, 0])
        )
        trackball.down((320, 240))
        trackball.drag((321, 240))
        pose1 = trackball.pose
        trackball.drag((320, 240))
        trackball.resize((1280, 960))
        trackball.down((640, 480))
        trackball.drag((642, 480))
        pose2 = trackball.pose
        g.np.testing.assert_allclose(pose1, pose2)

    def test_drag_rotate(self):
        trackball = Trackball(
            pose=g.np.eye(4),
            size=(640, 480),
            scale=1.0,
            target=g.np.array([0, 0, 0])
        )

        # rotates around y-axis
        trackball.set_state(Trackball.STATE_ROTATE)
        trackball.down((320, 240))
        trackball.drag((321, 240))
        assert trackball.pose[0, 0] < 1
        assert trackball.pose[0, 1] == 0
        assert trackball.pose[0, 2] < 0
        assert trackball.pose[1, 0] == 0
        assert trackball.pose[1, 1] == 1
        assert trackball.pose[1, 2] == 0
        assert trackball.pose[2, 0] > 0
        assert trackball.pose[2, 1] == 0
        assert trackball.pose[2, 2] < 1
        g.np.testing.assert_allclose(trackball.pose[:, 3], [0, 0, 0, 1])
        g.np.testing.assert_allclose(trackball.pose[3, :], [0, 0, 0, 1])
        # rotate back
        trackball.drag((320, 240))
        g.np.testing.assert_allclose(trackball.pose, g.np.eye(4))

        # rotate around x-axis
        trackball.drag((320, 241))
        assert trackball.pose[0, 0] == 1
        assert trackball.pose[0, 1] == 0
        assert trackball.pose[0, 2] == 0
        assert trackball.pose[1, 0] == 0
        assert trackball.pose[1, 1] < 1
        assert trackball.pose[1, 2] < 0
        assert trackball.pose[2, 0] == 0
        assert trackball.pose[2, 1] > 0
        assert trackball.pose[2, 2] < 1
        g.np.testing.assert_allclose(trackball.pose[:, 3], [0, 0, 0, 1])
        g.np.testing.assert_allclose(trackball.pose[3, :], [0, 0, 0, 1])
        # rotate back
        trackball.drag((320, 240))
        g.np.testing.assert_allclose(trackball.pose, g.np.eye(4))

    def test_drag_roll(self):
        trackball = Trackball(
            pose=g.np.eye(4),
            size=(640, 480),
            scale=1.0,
            target=g.np.array([0, 0, 0])
        )

        # rotates around z-axis
        trackball.set_state(Trackball.STATE_ROLL)
        trackball.down((321, 241))
        trackball.drag((320, 241))
        assert trackball.pose[0, 0] < 1
        assert trackball.pose[0, 1] > 0
        assert trackball.pose[0, 2] == 0
        assert trackball.pose[1, 0] < 0
        assert trackball.pose[1, 1] < 1
        assert trackball.pose[1, 2] == 0
        assert trackball.pose[2, 0] == 0
        assert trackball.pose[2, 1] == 0
        assert trackball.pose[2, 2] == 1
        g.np.testing.assert_allclose(trackball.pose[:, 3], [0, 0, 0, 1])
        g.np.testing.assert_allclose(trackball.pose[3, :], [0, 0, 0, 1])
        # rotate back
        trackball.drag((321, 241))
        g.np.testing.assert_allclose(trackball.pose, g.np.eye(4))

    def test_drag_pan(self):
        trackball = Trackball(
            pose=g.np.eye(4),
            size=(640, 480),
            scale=1.0,
            target=g.np.array([0, 0, 0])
        )

        # translate to x
        trackball.set_state(Trackball.STATE_PAN)
        trackball.down((321, 241))
        trackball.drag((320, 241))
        g.np.testing.assert_allclose(trackball.pose[:3, :3], g.np.eye(3))
        assert trackball.pose[0, 3] != 0
        assert trackball.pose[1, 3] == 0
        assert trackball.pose[2, 3] == 0

    def test_drag_zoom(self):
        pose = g.trimesh.transformations.translation_matrix([0, 0, 1])
        trackball = Trackball(
            pose=pose,
            size=(640, 480),
            scale=1.0,
            target=g.np.array([0, 0, 0])
        )

        # translate to x
        trackball.set_state(Trackball.STATE_ZOOM)
        trackball.down((320, 240))
        trackball.drag((320, 241))
        g.np.testing.assert_allclose(trackball.pose[:3, :3], g.np.eye(3))
        assert trackball.pose[3, 0] == 0
        assert trackball.pose[3, 1] == 0
        assert trackball.pose[3, 2] < 1
        g.np.testing.assert_allclose(trackball.pose[3, :], [0, 0, 0, 1])

    def test_scroll(self):
        pose = g.trimesh.transformations.translation_matrix([0, 0, 1])
        trackball = Trackball(
            pose=pose,
            size=(640, 480),
            scale=1.0,
            target=g.np.array([0, 0, 0])
        )
        g.np.testing.assert_allclose(trackball.pose[:3, :3], g.np.eye(3))
        g.np.testing.assert_allclose(trackball.pose[:3, 3], [0, 0, 1])
        trackball.scroll(1)
        g.np.testing.assert_allclose(trackball.pose[:3, :3], g.np.eye(3))
        g.np.testing.assert_allclose(trackball.pose[:3, 3], [0, 0, 0.9])

    def test_rotate(self):
        trackball = Trackball(
            pose=g.np.eye(4),
            size=(640, 480),
            scale=1.0,
            target=g.np.array([0, 0, 0])
        )
        # rotates around y-axis
        trackball.rotate(g.np.deg2rad(1))
        assert trackball.pose[0, 0] < 1
        assert trackball.pose[0, 1] == 0
        assert trackball.pose[0, 2] > 0
        assert trackball.pose[1, 0] == 0
        assert trackball.pose[1, 1] == 1
        assert trackball.pose[1, 2] == 0
        assert trackball.pose[2, 0] < 0
        assert trackball.pose[2, 1] == 0
        assert trackball.pose[2, 2] < 1
        g.np.testing.assert_allclose(trackball.pose[:, 3], [0, 0, 0, 1])
        g.np.testing.assert_allclose(trackball.pose[3, :], [0, 0, 0, 1])


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
Numpy encode/decode/utility implementations for run length encodings.

# Run Length Encoded Features

Encoding/decoding functions for run length encoded data.

We include code for two variations:

* run length encoding (RLE)
* binary run length encdoing (BRLE)

RLE stores sequences of repeated values as the value followed by its count, e.g.

```python
dense_to_rle([5, 5, 3, 2, 2, 2, 2, 6]) == [5, 2, 3, 1, 2, 4, 6, 1]
```

i.e. the value `5` is repeated `2` times, then `3` is repeated `1` time, `2` is
repeated `4` times and `6` is repeated `1` time.

BRLE is an optimized form for when the stored values can only be `0` or `1`.
This means we only need to save the counts, and assume the values alternate
(starting at `0`).

```python
dense_to_brle([1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0]) == \
              [0, 2, 4, 7, 2]
```

i.e. the value zero occurs `0` times, followed by `2` ones, `4` zeros, `7` ones
and `2` zeros.

Sequences with counts exceeding the data type's maximum value have to be
handled carefully. For example, the `uint8` encoding of 300 zeros
(`uint8` has a max value of 255) is:

* RLE: `[0, 255, 0, 45]`  (`0` repeated `255` times + `0` repeated `45` times)
* BRLE: `[255, 0, 45, 0]` (`255` zeros + `0` ones + `45` zeros + `0` ones)

This module contains implementations of various RLE/BRLE operations.
"""
import functools
import numpy as np


def brle_length(brle):
    """Optimized implementation of `len(brle_to_dense(brle))`"""
    return np.sum(brle)


def rle_length(rle):
    """Optimized implementation of `len(rle_to_dense(rle_to_brle(rle)))`"""
    return np.sum(rle[1::2])


def rle_to_brle(rle, dtype=None):
    """
    Convert run length encoded (RLE) value/counts to BRLE.

    RLE data is stored in a rank 1 array with each pair giving:
    (value, count)

    e.g. the RLE encoding of [4, 4, 4, 1, 1, 6] is [4, 3, 1, 2, 6, 1].

    Parameters
    ----------
    rle : (n,) int
      Run length encoded data

    Returns
    ----------
    equivalent binary run length encoding. a list if dtype is None,
        otherwise brle_to_brle is called on that list before returning.

    Raises
    ----------
    ValueError
      If any of the even counts of `rle` are not zero or 1.
    """
    curr_val = 0
    out = [0]
    acc = 0
    for value, count in np.reshape(rle, (-1, 2)):
        acc += count
        if value not in (0, 1):
            raise ValueError(
                "Invalid run length encoding for conversion to BRLE")
        if value == curr_val:
            out[-1] += count
        else:
            out.append(int(count))
            curr_val = value
    if len(out) % 2:
        out.append(0)
    if dtype is not None:
        out = brle_to_brle(out, dtype=dtype)
    return out


def brle_logical_not(brle):
    """
    Get the BRLE encoding of the `logical_not`ed dense form of `brle`.

    Equivalent to `dense_to_brle(np.logical_not(brle_to_dense(brle)))` but
    highly optimized - just pads brle with a 0 on each end (or strips is
    existing endpoints are both zero).

    Parameters
    ----------
    brle: rank 1 int array of binary run length encoded data

    Returns
    ----------
    rank 1 int array of binary run length encoded data corresponding to
    element-wise not of the input.
    """
    if brle[0] or brle[-1]:
        return np.pad(brle, [1, 1], mode='constant')
    else:
        return brle[1:-1]


def merge_brle_lengths(lengths):
    """Inverse of split_long_brle_lengths."""
    if len(lengths) == 0:
        return []

    out = [int(lengths[0])]
    accumulating = False
    for length in lengths[1:]:
        if accumulating:
            out[-1] += length
            accumulating = False
        else:
            if length == 0:
                accumulating = True
            else:
                out.append(int(length))
    return out


def split_long_brle_lengths(lengths, dtype=np.int64):
    """
    Split lengths that exceed max dtype value.

    Lengths `l` are converted into [max_val, 0] * l // max_val + [l % max_val]

    e.g. for dtype=np.uint8 (max_value == 255)
    ```
    split_long_brle_lengths([600, 300, 2, 6], np.uint8) == \
             [255, 0, 255, 0, 90, 255, 0, 45, 2, 6]
    ```
    """
    lengths = np.asarray(lengths)
    max_val = np.iinfo(dtype).max
    bad_length_mask = lengths > max_val
    if np.any(bad_length_mask):
        # there are some bad lengths
        nl = len(lengths)
        repeats = np.asarray(lengths) // max_val
        remainders = (lengths % max_val).astype(dtype)

        lengths = np.concatenate(
            [np.array([max_val, 0] * repeat + [remainder], dtype=dtype)
             for repeat, remainder in zip(repeats, remainders)])
        lengths = lengths.reshape((np.sum(repeats) * 2 + nl,)).astype(dtype)
        return lengths
    elif lengths.dtype != dtype:
        return lengths.astype(dtype)
    else:
        return lengths


def dense_to_brle(dense_data, dtype=np.int64):
    """
    Get the binary run length encoding of `dense_data`.

    Parameters
    ----------
    dense_data: rank 1 bool array of data to encode.
    dtype: numpy int type.

    Returns
    ----------
    Binary run length encoded rank 1 array of dtype `dtype`.

    Raises
    ----------
    ValuError if dense_data is not a rank 1 bool array.
    """
    if dense_data.dtype != bool:
        raise ValueError("`dense_data` must be bool")
    if len(dense_data.shape) != 1:
        raise ValueError("`dense_data` must be rank 1.")
    n = len(dense_data)
    starts = np.r_[0, np.flatnonzero(dense_data[1:] != dense_data[:-1]) + 1]
    lengths = np.diff(np.r_[starts, n])
    lengths = split_long_brle_lengths(lengths, dtype=dtype)
    if dense_data[0]:
        lengths = np.pad(lengths, [1, 0], mode='constant')
    return lengths


_ft = np.array([False, True], dtype=bool)


def brle_to_dense(brle_data, vals=None):
    """Decode binary run length encoded data to dense.

    Parameters
    ----------
    brle_data: BRLE counts of False/True values
    vals: if not `None`, a length 2 array/list/tuple with False/True substitute
        values, e.g. brle_to_dense([2, 3, 1, 0], [7, 9]) == [7, 7, 9, 9, 9, 7]

    Returns
    ----------
    rank 1 dense data of dtype `bool if vals is None else vals.dtype`

    Raises
    ----------
    ValueError if vals it not None and shape is not (2,)
    """
    if vals is None:
        vals = _ft
    else:
        vals = np.asarray(vals)
        if vals.shape != (2,):
            raise ValueError("vals.shape must be (2,), got %s" % (vals.shape))
    ft = np.repeat(
        _ft[np.newaxis, :], (len(brle_data) + 1) // 2, axis=0).flatten()
    return np.repeat(ft[:len(brle_data)], brle_data).flatten()


def rle_to_dense(rle_data, dtype=np.int64):
    """Get the dense decoding of the associated run length encoded data."""
    values, counts = np.split(np.reshape(rle_data, (-1, 2)), 2, axis=-1)
    if dtype is not None:
        values = np.asanyarray(values, dtype=dtype)
    try:
        result = np.repeat(np.squeeze(values, axis=-1),
                           np.squeeze(counts, axis=-1))
    except TypeError:
        # on windows it sometimes fails to cast data type
        result = np.repeat(np.squeeze(values.astype(np.int64), axis=-1),
                           np.squeeze(counts.astype(np.int64), axis=-1))
    return result


def dense_to_rle(dense_data, dtype=np.int64):
    """Get run length encoding of the provided dense data."""
    n = len(dense_data)
    starts = np.r_[0, np.flatnonzero(dense_data[1:] != dense_data[:-1]) + 1]
    lengths = np.diff(np.r_[starts, n])
    values = dense_data[starts]
    values, lengths = split_long_rle_lengths(values, lengths, dtype=dtype)
    out = np.stack((values, lengths), axis=1)
    return out.flatten()


def split_long_rle_lengths(values, lengths, dtype=np.int64):
    """
    Split long lengths in the associated run length encoding.

    e.g.
    ```python
    split_long_rle_lengths([5, 300, 2, 12], np.uint8) == [5, 255, 5, 45, 2, 12]
    ```

    Parameters
    ----------
    values: values column of run length encoding, or `rle[::2]`
    lengths: counts in run length encoding, or `rle[1::2]`
    dtype: numpy data type indicating the maximum value.

    Returns
    ----------
    values, lengths associated with the appropriate splits. `lengths` will be
    of type `dtype`, while `values` will be the same as the value passed in.
    """
    max_length = np.iinfo(dtype).max
    lengths = np.asarray(lengths)
    repeats = lengths // max_length
    if np.any(repeats):
        repeats += 1
        remainder = lengths % max_length
        values = np.repeat(values, repeats)
        lengths = np.empty(len(repeats), dtype=dtype)
        lengths.fill(max_length)
        lengths = np.repeat(lengths, repeats)
        lengths[np.cumsum(repeats) - 1] = remainder
    elif lengths.dtype != dtype:
        lengths = lengths.astype(dtype)
    return values, lengths


def merge_rle_lengths(values, lengths):
    """Inverse of split_long_rle_lengths except returns normal python lists."""
    ret_values = []
    ret_lengths = []
    curr = None
    for value, length in zip(values, lengths):
        if length == 0:
            continue
        if value == curr:
            ret_lengths[-1] += length
        else:
            curr = value
            ret_lengths.append(int(length))
            ret_values.append(value)
    return ret_values, ret_lengths


def brle_to_rle(brle, dtype=np.int64):
    if len(brle) % 2 == 1:
        brle = np.concatenate([brle, [0]])
    lengths = brle
    values = np.tile(_ft, len(brle) // 2)
    return rle_to_rle(
        np.stack((values, lengths), axis=1).flatten(), dtype=dtype)


def brle_to_brle(brle, dtype=np.int64):
    """
    Almost the identity function.

    Checks for possible merges and required splits.
    """
    return split_long_brle_lengths(merge_brle_lengths(brle), dtype=dtype)


def rle_to_rle(rle, dtype=np.int64):
    """
    Almost the identity function.

    Checks for possible merges and required splits.
    """
    values, lengths = np.reshape(rle, (-1, 2)).T
    values, lengths = merge_rle_lengths(values, lengths)
    values, lengths = split_long_rle_lengths(values, lengths, dtype=dtype)
    return np.stack((values, lengths), axis=1).flatten()


def _unsorted_gatherer(indices, sorted_gather_fn):
    if not isinstance(indices, np.ndarray):
        indices = np.array(indices, copy=False)
    order = np.argsort(indices)
    ordered_indices = indices[order]

    def f(data, dtype=None):
        result = np.empty(
            len(order), dtype=dtype or getattr(
                data, 'dtype', None))
        result[order] = tuple(sorted_gather_fn(data, ordered_indices))
        return result

    return f


def sorted_rle_gather_1d(rle_data, ordered_indices):
    """
    Gather brle_data at ordered_indices.

    This is equivalent to `rle_to_dense(brle_data)[ordered_indices]` but avoids
    the decoding.

    Parameters
    ----------
    brle_data: iterable of run-length-encoded data.
    ordered_indices: iterable of ints in ascending order.

    Returns
    ----------
    `brle_data` iterable of values at the dense indices, same length as
    ordered indices.
    """
    data_iter = iter(rle_data)
    index_iter = iter(ordered_indices)
    try:
        index = next(index_iter)
    except StopIteration:
        return
    start = 0
    while True:
        while start <= index:
            try:
                value = next(data_iter)
                start += next(data_iter)
            except StopIteration:
                raise IndexError(
                    'Index %d out of range of raw_values length %d'
                    % (index, start))
        try:
            while index < start:
                yield value
                index = next(index_iter)
        except StopIteration:
            break


def rle_mask(rle_data, mask):
    """
    Perform masking of the input run-length data.

    Parameters
    ----------
    rle_data: iterable of run length encoded data
    mask: iterable of bools corresponding to the dense mask.

    Returns
    ----------
    iterable of dense values of rle_data wherever mask is True.
    """
    data_iter = iter(rle_data)
    mask_iter = iter(mask)
    while True:
        try:
            value = next(data_iter)
            count = next(data_iter)
        except StopIteration:
            break
        for _ in range(count):
            m = next(mask_iter)
            if m:
                yield value


def brle_mask(rle_data, mask):
    """
    Perform masking of the input binary run-length data.

    Parameters
    ----------
    brle_data: iterable of binary run length encoded data
    mask: iterable of bools corresponding to the dense mask.

    Returns
    ----------
    iterable dense values of brle_data wherever mask is True.
    """
    data_iter = iter(rle_data)
    mask_iter = iter(mask)
    value = True
    while True:
        try:
            value = not value
            count = next(data_iter)
        except StopIteration:
            break
        for _ in range(count):
            m = next(mask_iter)
            if m:
                yield value


def rle_gatherer_1d(indices):
    """
    Get a gather function at the given indices.

    Because gathering on RLE data requires sorting, for instances where
    gathering at the same indices on different RLE data this can save the
    sorting process.

    If only gathering on a single RLE iterable, use `rle_gather_1d`.

    Parameters
    ----------
    indices: iterable of integers

    Returns
    ----------
    gather function, mapping `(rle_data, dtype=None) -> values`.
    `values` will have the same length as `indices` and dtype provided,
    or rle_data.dtype if no dtype is provided.
    """
    return _unsorted_gatherer(indices, sorted_rle_gather_1d)


def rle_gather_1d(rle_data, indices, dtype=None):
    """
    Gather RLE data values at the provided dense indices.

    This is equivalent to `rle_to_dense(rle_data)[indices]` but the
    implementation does not require the construction of the dense array.

    If indices is known to be in order, use `sorted_gather_1d`.

    Parameters
    ----------
    rle_data: run length encoded data
    indices: dense indices
    dtype: numpy dtype. If not provided, uses rle_data.dtype

    Returns
    ----------
    numpy array, dense data at indices, same length as indices and dtype as
    rle_data
    """
    return rle_gatherer_1d(indices)(rle_data, dtype=dtype)


def sorted_brle_gather_1d(brle_data, ordered_indices):
    """
    Gather brle_data at ordered_indices.

    This is equivalent to `brle_to_dense(brle_data)[ordered_indices]` but
    avoids the decoding.

    Parameters
    ----------
    raw_data: iterable of run-length-encoded data.
    ordered_indices: iterable of ints in ascending order.

    Returns
    ----------
    `raw_data` iterable of values at the dense indices, same length as
    ordered indices.
    """
    data_iter = iter(brle_data)
    index_iter = iter(ordered_indices)
    try:
        index = next(index_iter)
    except StopIteration:
        return
    start = 0
    value = True
    while True:
        while start <= index:
            try:
                value = not value
                start += next(data_iter)
            except StopIteration:
                raise IndexError(
                    'Index %d out of range of raw_values length %d'
                    % (index, start))
        try:
            while index < start:
                yield value
                index = next(index_iter)
        except StopIteration:
            break


def brle_gatherer_1d(indices):
    """
    Get a gather function at the given indices.

    Because gathering on BRLE data requires sorting, for instances where
    gathering at the same indices on different RLE data this can save the
    sorting process.

    If only gathering on a single RLE iterable, use `brle_gather_1d`.

    Parameters
    ----------
    indices: iterable of integers

    Returns
    ----------
    gather function, mapping `(rle_data, dtype=None) -> values`.
    `values` will have the same length as `indices` and dtype provided,
    or rle_data.dtype if no dtype is provided.
    """
    return functools.partial(
        _unsorted_gatherer(indices, sorted_brle_gather_1d), dtype=bool)


def brle_gather_1d(brle_data, indices):
    """
    Gather BRLE data values at the provided dense indices.

    This is equivalent to `rle_to_dense(rle_data)[indices]` but the
    implementation does not require the construction of the dense array.

    If indices is known to be in order, use `sorted_brle_gather_1d`.

    Parameters
    ----------
    rle_data: run length encoded data
    indices: dense indices

    Returns
    ----------
    numpy array, dense data at indices, same length as indices and dtype as
    rle_data
    """
    return brle_gatherer_1d(indices)(brle_data)


def brle_reverse(brle_data):
    """Equivalent to dense_to_brle(brle_to_dense(brle_data)[-1::-1])."""
    if len(brle_data) % 2 == 0:
        brle_data = np.concatenate([brle_data, [0]], axis=0)
    end = -1 if brle_data[-1] == 0 else None
    return brle_data[-1:end:-1]


def rle_reverse(rle_data):
    """Get the rle encoding of the reversed dense array."""
    if not isinstance(rle_data, np.ndarray):
        rle_data = np.array(rle_data, copy=False)
    rle_data = np.reshape(rle_data, (-1, 2))
    rle_data = rle_data[-1::-1]
    return np.reshape(rle_data, (-1,))


def rle_to_sparse(rle_data):
    """Get dense indices associated with non-zeros."""
    indices = []
    values = []
    it = iter(rle_data)
    index = 0
    try:
        while True:
            value = next(it)
            counts = next(it)
            end = index + counts
            if value:
                indices.append(np.arange(index, end, dtype=np.int64))
                values.append(np.repeat(value, counts))
            index = end
    except StopIteration:
        pass
    if len(indices) == 0:
        assert len(values) == 0
        return indices, values

    indices = np.concatenate(indices)
    values = np.concatenate(values)
    return indices, values


def brle_to_sparse(brle_data, dtype=np.int64):
    ends = np.cumsum(brle_data)
    indices = [np.arange(s, e, dtype=dtype) for s, e in
               zip(ends[::2], ends[1::2])]
    return np.concatenate(indices)


def rle_strip(rle_data):
    """
    Remove leading and trailing zeros.

    Parameters
    ----------
    rle_data: run length encoded data

    Returns
    ----------
    (stripped_rle_data, padding)
        stripped_rle_data: rle data without any leading or trailing zeros
        padding: 2-element dense padding
    """
    rle_data = np.reshape(rle_data, (-1, 2))
    start = 0
    final_i = len(rle_data)
    for i, (val, count) in enumerate(rle_data):
        if val and count > 0:
            final_i = i
            break
        else:
            start += count

    end = 0
    final_j = len(rle_data)
    for j, (val, count) in enumerate(rle_data[::-1]):
        if val and count > 0:
            final_j = j
            break
        else:
            end += count

    rle_data = rle_data[
        final_i:None if final_j == 0 else -final_j].reshape((-1,))
    return rle_data, (start, end)


def brle_strip(brle_data):
    """
    Remove leading and trailing zeros.

    Parameters
    ----------
    brle_data: binary run length encoded data.

    Returns
    ----------
    (stripped_brle_data, padding)
    stripped_brle_data: rle data without any leading or trailing zeros
    padding: 2-element dense padding
    """
    start = 0
    val = True
    final_i = len(brle_data)
    for i, count in enumerate(brle_data):
        val = not val
        if val and count > 0:
            final_i = i
            break
        else:
            start += count
    end = 0
    final_j = len(brle_data)
    val = bool(len(brle_data) % 2)
    for j, count in enumerate(brle_data[::-1]):
        val = not val
        if val and count > 0:
            final_j = j
            break
        else:
            end += count

    brle_data = brle_data[final_i:None if final_j == 0 else -final_j]
    brle_data = np.concatenate([[0], brle_data])
    return brle_data, (start, end)



"""
objects.py
--------------

Deal with objects which hold visual properties, like
ColorVisuals and TextureVisuals.
"""
import numpy as np

from .material import pack
from .texture import TextureVisuals
from .color import ColorVisuals, color_to_uv


def create_visual(**kwargs):
    """
    Create Visuals object from keyword arguments.

    Parameters
    -----------
    face_colors : (n, 3|4) uint8
      Face colors
    vertex_colors : (n, 3|4) uint8
      Vertex colors
    mesh : trimesh.Trimesh
      Mesh object

    Returns
    ----------
    visuals : ColorVisuals
      Visual object created from arguments
    """
    return ColorVisuals(**kwargs)


def concatenate(visuals, *args):
    """
    Concatenate multiple visual objects.

    Parameters
    ----------
    visuals : ColorVisuals or list
      Visuals to concatenate
    *args :  ColorVisuals or list
      More visuals to concatenate

    Returns
    ----------
    concat : Visuals
      If all are color
    """
    # get a flat list of Visuals objects
    if len(args) > 0:
        visuals = np.append(visuals, args)
    else:
        visuals = np.array(visuals)

    # if there are any texture visuals convert all to texture
    if any(v.kind == 'texture' for v in visuals):
        # first collect materials and UV coordinates
        mats = []
        uvs = []
        for v in visuals:
            if v.kind == 'texture':
                mats.append(v.material)
                if v.uv is None:
                    # otherwise use zeros
                    uvs.append(np.zeros((len(v.mesh.vertices), 2)) + 0.5)
                else:
                    # if uvs are of correct shape use them
                    uvs.append(v.uv)

            else:
                # create a material and UV coordinates from vertex colors
                color_mat, color_uv = color_to_uv(
                    vertex_colors=v.vertex_colors)
                mats.append(color_mat)
                uvs.append(color_uv)
        # pack the materials and UV coordinates into one
        new_mat, new_uv = pack(materials=mats, uvs=uvs)
        return TextureVisuals(material=new_mat, uv=new_uv)

    # convert all visuals to the kind of the first
    kind = visuals[0].kind
    if kind == 'face':
        colors = np.vstack([
            v.face_colors for v in visuals])
        return ColorVisuals(face_colors=colors)
    elif kind == 'vertex':
        colors = np.vstack([
            v.vertex_colors for v in visuals])
        return ColorVisuals(vertex_colors=colors)

    return ColorVisuals()



try:
    from . import generic as g
except BaseException:
    import generic as g


# tetToTris extracts outward facing triangles
# of a tetrahedron given in this order
#    1
#  / | \
# 0-----2
#  \ | /
#    3

triIdxs = g.np.ravel([[2, 1, 0], [3, 0, 1], [3, 2, 0], [3, 1, 2]])


def tetToTris(tet):
    return g.np.reshape(tet[triIdxs], (-1, 3))


class InertiaTest(g.unittest.TestCase):

    def test_inertia(self):
        t0 = g.np.array([[-0.419575686853, -0.898655215203, -0.127965023308, 0.],
                         [0.712589964872, -0.413418145015, 0.566834172697, 0.],
                         [-0.562291548012, 0.146643245877, 0.813832890385, 0.],
                         [0., 0., 0., 1.]])
        t1 = g.np.array([[0.343159553585, 0.624765521319, -0.701362648103, 0.],
                         [0.509982849005, -0.750986657709, -0.419447891476, 0.],
                         [-0.788770571525, -0.213745370274, -0.57632794673, 0.],
                         [0., 0., 0., 1.]])

        # make sure our transformations are actually still transformations
        assert g.np.abs(g.np.dot(t0, t0.T) - g.np.eye(4)).max() < 1e-10
        assert g.np.abs(g.np.dot(t1, t1.T) - g.np.eye(4)).max() < 1e-10

        c = g.trimesh.primitives.Cylinder(height=10,
                                          radius=1,
                                          sections=720,  # number of slices
                                          transform=t0)
        c0m = c.moment_inertia.copy()
        c0 = g.trimesh.inertia.cylinder_inertia(c.volume,
                                                c.primitive.radius,
                                                c.primitive.height,
                                                c.primitive.transform)

        ct = g.np.abs((c0m / c0) - 1)

        # we are comparing an inertia tensor from a mesh of a cylinder
        # to an inertia tensor from an actual cylinder, so allow for some
        # discretization uncertainty
        assert ct.max() < 1e-3

        # check our principal axis calculation against this cylinder
        # the direction (long axis) of the cylinder should correspond to
        # the smallest principal component of inertia, AKA rotation along
        # the axis, rather than the other two which are perpendicular
        components, vectors = g.trimesh.inertia.principal_axis(
            c.moment_inertia)
        # inferred cylinder axis
        inferred = vectors[components.argmin()]

        # inferred cylinder axis should be close to actual cylinder axis
        axis_test = g.np.allclose(g.np.abs(inferred),
                                  g.np.abs(c.direction))
        assert axis_test

        # make sure Trimesh attribute is plumbed correctly
        assert g.np.allclose(c.principal_inertia_components, components)
        assert g.np.allclose(c.principal_inertia_vectors, vectors)

        # the other two axis of the cylinder should be identical
        assert g.np.abs(g.np.diff(g.np.sort(components)[-2:])).max() < 1e-8

        m = g.get_mesh('featuretype.STL')
        i0 = m.moment_inertia.copy()
        # rotate the moment of inertia
        i1 = g.trimesh.inertia.transform_inertia(
            transform=t0, inertia_tensor=i0)

        # rotate the mesh
        m.apply_transform(t0)
        # check to see if the rotated mesh + recomputed moment of inertia
        # is close to the rotated moment of inertia
        tf_test = g.np.abs((m.moment_inertia / i1) - 1)
        assert tf_test.max() < 1e-6

        # do it again with another transform
        i2 = g.trimesh.inertia.transform_inertia(
            transform=t1, inertia_tensor=i1)
        m.apply_transform(t1)
        tf_test = g.np.abs((m.moment_inertia / i2) - 1)
        assert tf_test.max() < 1e-6

    def test_primitives(self):
        primitives = [g.trimesh.primitives.Cylinder(height=5),
                      g.trimesh.primitives.Box(),
                      g.trimesh.primitives.Sphere(radius=1.23)]
        for p in primitives:
            for _i in range(100):
                # check to make sure the analytic inertia tensors are relatively
                # close to the meshed inertia tensor (order of magnitude and
                # sign)
                b = p.to_mesh()
                comparison = g.np.abs(
                    p.moment_inertia - b.moment_inertia)
                c_max = comparison.max() / g.np.abs(p.moment_inertia).max()
                assert c_max < .1

                if hasattr(p.primitive, 'transform'):
                    matrix = g.trimesh.transformations.random_rotation_matrix()
                    p.primitive.transform = matrix
                elif hasattr(p.primitive, 'center'):
                    p.primitive.center = g.random(3)

    def test_tetrahedron(self):
        # Based on the 'numerical example' of the paper:

        # Explicit Exact Formulas for the 3-D Tetrahedron Inertia Tensor
        # in Terms of its Vertex Coordinates, [F. Tonon, 2004]
        # http://thescipub.com/pdf/jmssp.2005.8.11.pdf

        # set up given vertices
        vertices = g.np.float32([[8.3322, -11.86875, 0.93355],
                                 [0.75523, 5., 16.37072],
                                 [52.61236, 5., -5.3858],
                                 [2., 5., 3.]])

        # set up a simple trimesh tetrahedron
        tris = tetToTris(g.np.int32([0, 1, 2, 3]))
        tm_tet = g.trimesh.Trimesh(vertices, tris)

        # 'ground truth' values from the paper
        # however, there are some minor flaws

        # a small deviation in the last decimal of the mass-centers' x:
        CM_gt = g.np.float32([15.92492, 0.78281, 3.72962])

        # moment of inertia values
        a_mi, b_mi, c_mi = [43520.33257, 194711.28938, 191168.76173]
        # principle inertia values
        a_pi, b_pi, c_pi = [4417.6615, -46343.16662, 11996.20119]

        # NOTE: I guess there is a mistake in the paper
        # b' (Eq. 9e) computes from x and z values
        # c' (Eq. 9f) computes from x and y values
        # therefore the given matrix E_Q (Eq. 2) is not correct
        # b' and c' need to be swapped like this:
        MI_gt = g.np.float32([[a_mi, -c_pi, -b_pi],
                              [-c_pi, b_mi, -a_pi],
                              [-b_pi, -a_pi, c_mi]])

        # check center of mass
        assert g.np.allclose(CM_gt, tm_tet.center_mass)

        # check moment of inertia tensor
        assert g.np.allclose(MI_gt, tm_tet.moment_inertia)

    def test_cube_with_tetras(self):

        # set up a unit cube, vertices in this order:
        #   1-----2
        #  /|    /|
        # 0-+---3 |
        # | 5---+-6
        # |/    |/
        # 4-----7

        vertices = g.np.float32([[-1, -1, 1],
                                 [-1, 1, 1],
                                 [1, 1, 1],
                                 [1, -1, 1],
                                 [-1, -1, -1],
                                 [-1, 1, -1],
                                 [1, 1, -1],
                                 [1, -1, -1]]) * 0.5

        # 6 quad faces for the cube
        quads = g.np.int32([[3, 2, 1, 0],
                            [0, 1, 5, 4],
                            [1, 2, 6, 5],
                            [2, 3, 7, 6],
                            [3, 0, 4, 7],
                            [4, 5, 6, 7]])

        # set up two different tetraherdalizations of a cube
        # using 5 tets (1 in the middle and 4 around)
        tetsA = g.np.int32([[0, 1, 3, 4],
                            [1, 2, 3, 6],
                            [1, 4, 5, 6],
                            [3, 4, 6, 7],
                            [1, 3, 4, 6]])
        # using 6 sets (around the cube diagonal)
        tetsB = g.np.int32([[0, 1, 2, 6],
                            [0, 1, 6, 5],
                            [0, 4, 5, 6],
                            [0, 4, 6, 7],
                            [0, 3, 7, 6],
                            [0, 2, 3, 6]])

        # create a trimesh cube from vertices and faces
        tm_cube = g.trimesh.Trimesh(vertices, quads)

        # https://en.wikipedia.org/wiki/List_of_moments_of_inertia
        # ground truth for a unit cube with side length s = mass m = 1
        # I = 1/6 * m * s^2
        MI_gt = g.np.eye(3) / 6
        assert g.np.allclose(MI_gt, tm_cube.moment_inertia)

        # compare both tetrahedralizations
        # should be equivalent to each other and to the cube
        for tets in [tetsA, tetsB]:
            # create trimesh tets from vertices and triangles
            tm_tets = [g.trimesh.Trimesh(vertices, tetToTris(t)) for t in tets]

            # get mass, center of mass, and moment of inertia for each tet
            Ms = [tm_tet.mass for tm_tet in tm_tets]
            CMs = [tm_tet.center_mass for tm_tet in tm_tets]
            MIs = [tm_tet.moment_inertia for tm_tet in tm_tets]

            # compute total mass and center of mass
            mass = g.np.sum(Ms)
            center_mass = g.np.dot(Ms, CMs)

            # compare with unit cube
            assert g.np.abs(tm_cube.mass - mass) < 1e-6
            assert g.np.allclose(tm_cube.center_mass, center_mass)

            # the moment of inertia tensors for the individual tetrahedra
            # have to be re-assembled, using the parallel-axis-theorem
            # https://en.wikipedia.org/wiki/Parallel_axis_theorem#Tensor_generalization

            E = g.np.eye(3)
            MI = g.np.zeros((3, 3), g.np.float32)
            for i in range(len(tm_tets)):
                R = CMs[i] - center_mass
                MI += MIs[i] + Ms[i] * (g.np.dot(R, R) * E - g.np.outer(R, R))

            assert g.np.allclose(MI_gt, MI)

    def test_frame_inertia_box(self):
        # set up a rectangular cuboid, vertices as visible below.
        # The edges have the lengths 1,2 and 3
        #           ^ x0, x2
        #    y2 ^  /    b
        #       | 1-----------2
        #       |/|          /|
        #       0-+---------3-+----> y0, x1, z2
        #      /| |         | |
        #     / | |         | |
        # y1 v  | |         | |h
        #       | |         | |
        #       | |         | |
        #       | 5---------+-6
        #       |/          |/ d
        #       4-----------7
        #       |
        #       v z0, z1

        h = 3
        b = 2
        d = 1
        # definition in frame 0
        vertices = g.np.float32([[0, 0, 0],
                                 [d, 0, 0],
                                 [d, b, 0],
                                 [0, b, 0],
                                 [0, 0, h],
                                 [d, 0, h],
                                 [d, b, h],
                                 [0, b, h]])

        # 6 quad faces for the cube
        quads = g.np.int32([[3, 2, 1, 0],
                            [0, 1, 5, 4],
                            [1, 2, 6, 5],
                            [2, 3, 7, 6],
                            [3, 0, 4, 7],
                            [4, 5, 6, 7]])

        # create a trimesh cube from vertices and faces
        tm_cube = g.trimesh.Trimesh(vertices, quads)

        # set density of 1 and calculate mass
        tm_cube.density = 1
        mass = tm_cube.density * h * b * d
        assert tm_cube.mass == mass

        def parallel_axis_theorem(inertia, mass, a1, a2, a3):
            """
            Apply parallel axis theorem

            Parameters
            ------------
            inertia: (3,3) float
              inertia tensor around centre of mass
            mass : float
            a1, a2, a3: float
              difference between desired rotation center and center of mass in x,y,z

            Returns
            -----------
            rotated_inertia : (3,3) float
            Inertia tensor

            """
            # copy from wikipedia
            return inertia + mass * g.np.array(
                [[a2**2 + a3**2, -a1 * a2, -a1 * a3],
                 [-a1 * a2, a1**2 +
                  a3**2, -a2 * a3],
                 [-a1 * a3, -a2 * a3, a1**2 + a2**2]])

        # CHECK FRAME 0
        # analytical calculations of inertia tensor by hand
        inertia0 = 0.083333333333 * mass * g.np.diag([h**2 + b**2,
                                                      h**2 + d**2,
                                                      b**2 + d**2])
        a1 = - 0.5 * d
        a2 = - 0.5 * b
        a3 = - 0.5 * h
        inertia0 = parallel_axis_theorem(inertia0, mass, a1, a2, a3)
        # transformation from mesh base frame to frame 0
        t0 = g.np.eye(4)
        assert g.np.allclose(tm_cube.moment_inertia_frame(t0), inertia0)

        # CHECK FRAME 1
        # analytical calculations of inertia tensor by hand
        inertia1 = 0.083333333333 * mass * g.np.diag([h**2 + d**2,
                                                      h**2 + b**2,
                                                      b**2 + d**2])
        a1 = - 0.5 * b
        a2 = 0.5 * d
        a3 = - 0.5 * h
        inertia1 = parallel_axis_theorem(inertia1, mass, a1, a2, a3)
        # transformation from mesh base frame to frame 1
        # rotation of 90 deg around z-Axis
        t1 = g.trimesh.transformations.rotation_matrix(
            g.np.pi * 0.5, [0, 0, 1], [0, 0, 0])
        assert g.np.allclose(tm_cube.moment_inertia_frame(t1), inertia1)

        # CHECK FRAME 2
        # analytical calculations of inertia tensor by hand
        inertia2 = 0.083333333333 * mass * g.np.diag([h**2 + b**2,
                                                      b**2 + d**2,
                                                      h**2 + d**2])
        a1 = - 0.5 * d
        a2 = 0.5 * h
        a3 = - 0.5 * b
        inertia2 = parallel_axis_theorem(inertia2, mass, a1, a2, a3)
        # transformation from mesh base frame to frame 2
        # rotation of -90 deg around x-Axis
        t2 = g.trimesh.transformations.rotation_matrix(
            -g.np.pi * 0.5, [1, 0, 0], [0, 0, 0])
        assert g.np.allclose(tm_cube.moment_inertia_frame(t2), inertia2)

    def test_scene(self):
        # check our transforming of inertia matrices using the
        # parallel axis theorum by checking the inertia of a scene
        # with multiple instances transformed all over the scene
        # in two ways: by dumping it into a single mesh and then
        # running the heavily validated surface integral method,
        # and also running our "transform every matrix" logic
        # contained in `trimesh.inertia.scene_inertia`

        # scene with instancing
        s = g.get_mesh('cycloidal.3DXML')
        s._cache.clear()

        with g.Profiler() as P:
            ms = s.dump(concatenate=True)
            total_dump = ms.moment_inertia
        g.log.debug(P.output_text())

        # while we're at it check to make sure our composed
        # center mass, volume, and area calcs are all working
        assert g.np.allclose(s.center_mass, ms.center_mass)
        # dumped inertia should be the same as transformed inertia
        assert g.np.allclose(s.moment_inertia, ms.moment_inertia)
        assert g.np.allclose(s.volume, ms.volume)
        assert g.np.allclose(s.area, ms.area)

        # that produced the moment of inertia around the center of mass
        # of the scene, so for the `moment_inertia_frame` calculation
        # compare against that
        transform = g.tf.translation_matrix(ms.center_mass)
        s._cache.clear()

        with g.Profiler() as P:
            total_scene = g.trimesh.inertia.scene_inertia(
                s, transform=transform)
        g.log.debug(P.output_text())

        # compare the two calculation methods by percent
        diff = g.np.abs((total_dump - total_scene) / total_dump)
        assert diff.shape == (3, 3)

        # the two methods should return essentially identical results
        assert g.np.abs(diff).max() < 1e-3


class MassTests(g.unittest.TestCase):

    def setUp(self):
        # inertia numbers pulled from solidworks
        self.truth = g.data['mass_properties']
        self.meshes = dict()
        for data in self.truth:
            filename = data['filename']
            self.meshes[filename] = g.get_mesh(filename)

    def test_mass(self):

        for truth in self.truth:
            mesh = self.meshes[truth['filename']]
            calc = g.trimesh.triangles.mass_properties(
                triangles=mesh.triangles,
                density=truth['density'],
                skip_inertia=False)

            for key, _value in calc.items():
                if key not in truth:
                    continue
                if not g.np.allclose(calc[key], truth[key], atol=1e-2):
                    raise ValueError('{}({}):\n{}\n!=\n{}'.format(
                        truth['filename'],
                        key,
                        calc[key],
                        g.np.array(truth[key])))


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g

try:
    import collada
except ImportError:
    collada = None
except BaseException:
    # TODO : REMOVE WHEN UPSTREAM RELEASE FIXED
    # https://github.com/pycollada/pycollada/pull/92
    g.log.error('DAE fix not pushed yet!')
    collada = None


class DAETest(g.unittest.TestCase):
    def test_duck(self):
        """
        Load a collada scene with pycollada.
        """
        if collada is None:
            g.log.error('no pycollada to test!')
            return

        scene = g.get_mesh('duck.dae')

        assert len(scene.geometry) == 1
        assert len(scene.graph.nodes_geometry) == 1

    def test_shoulder(self):
        if collada is None:
            g.log.error('no pycollada to test!')
            return

        scene = g.get_mesh('shoulder.zae')
        assert len(scene.geometry) == 3
        assert len(scene.graph.nodes_geometry) == 3

    def test_export(self):
        if collada is None:
            g.log.error('no pycollada to test!')
            return

        a = g.get_mesh('ballA.off')
        r = a.export(file_type='dae')
        assert len(r) > 0

    def test_obj_roundtrip(self):
        # get a zipped-DAE scene
        s = g.get_mesh('duck.zae', force='mesh')
        with g.TemporaryDirectory() as root:
            # export using a file path so it can auto-create
            # a FilePathResolver to write the stupid assets
            path = g.os.path.join(root, 'duck.obj')
            s.export(path)
            # bring it back from outer space
            rec = g.trimesh.load(path, force='mesh')
        assert rec.visual.uv.ptp(axis=0).ptp() > 1e-5
        assert (s.visual.material.baseColorTexture.size ==
                rec.visual.material.image.size)

    def test_material_round(self):
        """
        Test to make sure materials survive a roundtrip
        with an actually identical result
        """
        s = g.get_mesh('blue_cube.dae')
        assert len(s.geometry) == 1
        m = next(iter(s.geometry.values()))

        rs = g.trimesh.load(
            file_obj=g.trimesh.util.wrap_as_stream(
                m.export(file_type='dae')),
            file_type='dae')
        assert len(rs.geometry) == 1
        r = next(iter(rs.geometry.values()))

        # this will compare everything in `material._data`
        assert (hash(m.visual.material) ==
                hash(r.visual.material))


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g
enc = g.trimesh.voxel.encoding
rl = g.trimesh.voxel.runlength
np = g.np

shape = (10, 10, 10)
dense_data = np.random.uniform(size=shape) < 0.2
rle = enc.RunLengthEncoding.from_dense(
    dense_data.reshape((-1,)), dtype=bool).reshape(shape)
brle = enc.BinaryRunLengthEncoding.from_dense(
    dense_data.reshape((-1,))).reshape(shape)
dense = enc.DenseEncoding(dense_data)
indices = np.column_stack(np.where(dense_data))
values = np.ones(shape=(indices.shape[0],), dtype=bool)
sparse = enc.SparseEncoding(indices, values, shape)

encodings = (
    dense,
    rle,
    brle,
    sparse,
)


class EncodingTest(g.unittest.TestCase):

    def _test_dense(self, encoding, data):
        np.testing.assert_equal(encoding.dense, data)

    def _test_rle(self, encoding, data):
        np.testing.assert_equal(
            encoding.run_length_data(), rl.dense_to_rle(data))

    def _test_brle(self, encoding, data):
        np.testing.assert_equal(
            encoding.binary_run_length_data(),
            rl.dense_to_brle(data))

    def _test_transpose(self, encoding, data, perm):
        encoding = encoding.transpose(perm)
        data = data.transpose(perm)
        self._test_dense(encoding, data)

    def _test_reshape(self, encoding, data, shape):
        encoding = encoding.reshape(shape)
        data = data.reshape(shape)
        self._test_dense(encoding, data)

    def _test_flat(self, encoding, data):
        self._test_dense(encoding.flat, data.reshape((-1,)))

    def _test_flipped(self, encoding, data, axes):
        if hasattr(axes, '__iter__'):
            for a in axes:
                data = np.flip(data, a)
        else:
            data = np.flip(data, axes)
        self._test_dense(encoding.flip(axes), data)

    def _test_composite(
            self, encoding, data, transpose=(0, 2, 1), reshape=(5, 2, -1),
            flatten=True, flip=(0, 2)):
        def check():
            self._test_dense(encoding, data)
        if transpose is not None:
            encoding = encoding.transpose(transpose)
            data = data.transpose(transpose)
            check()
        if reshape is not None:
            encoding = encoding.reshape(reshape)
            data = data.reshape(reshape)
            check()
        if flip:
            encoding = encoding.flip(flip)
            if hasattr(flip, '__iter__'):
                for a in flip:
                    data = np.flip(data, a)
            else:
                data = np.flip(data, flip)
            check()
        if flatten:
            encoding = encoding.flat
            data = data.reshape((-1,))
            check()
        self._test_dense(encoding, data)
        self._test_rle(encoding, data)
        self._test_brle(encoding, data)

    def test_dense(self):
        for encoding in encodings:
            self._test_dense(encoding, dense_data)

    def test_rle(self):
        for encoding in encodings:
            self._test_rle(encoding.flat, dense_data.reshape((-1,)))

    def test_brle(self):
        for encoding in encodings:
            self._test_brle(encoding.flat, dense_data.reshape((-1,)))

    def test_flipped(self):
        axes = (
            0, 1, 2,
            (0,),
            (0, 1),
            (0, 2),
            (1, 2),
            (0, 1, 2),
        )
        for encoding in encodings:
            for ax in axes:
                self._test_flipped(encoding, dense_data, ax)
            self.assertTrue(encoding.flip((0, 0)) is encoding)

    def test_transpose(self):
        perms = (
            (0, 2, 1),
            (0, 1, 2),
            (2, 1, 0),
        )
        for encoding in encodings:
            for perm in perms:
                self._test_transpose(encoding, dense_data, perm)
            perm = (0, 2, 1)
            if not isinstance(encoding, enc.DenseEncoding):
                self.assertTrue(
                    encoding.transpose(perm).transpose(perm) is encoding)

    def test_flat(self):
        for encoding in encodings:
            self._test_dense(encoding.flat, dense_data.reshape((-1,)))

    def test_reshape(self):
        shapes = (
            (10, 10, 10),
            (5, 20, 10),
            (50, 4, 5),
            (-1, 4, 5)
        )
        for encoding in encodings:
            for shape in shapes:
                self._test_dense(
                    encoding.reshape(shape), dense_data.reshape(shape))

    def test_composite(self):
        for encoding in encodings:
            self._test_composite(encoding, dense_data)

    def test_dense_stripped(self):
        base_shape = (5, 5, 5)
        dense = np.ones(base_shape, dtype=bool)
        padding = [[2, 2], [2, 2], [2, 2]]
        dense = np.pad(dense, padding, mode='constant')
        encoding = enc.DenseEncoding(dense)
        stripped, calculated_padding = encoding.stripped
        np.testing.assert_equal(calculated_padding, padding)
        np.testing.assert_equal(stripped.shape, base_shape)
        np.testing.assert_equal(stripped.dense, 1)

    def test_sparse_stripped(self):
        box = g.trimesh.primitives.Box()
        box.apply_translation([0.5, 0.5, 0.5])  # center at origin
        box.apply_scale(5)                      # 0 -> 5
        expected_sparse_indices = np.array(box.vertices)
        box.apply_translation([2, 2, 2])        # 2 -> 7
        sparse = np.array(box.vertices, dtype=int)
        encoding = enc.SparseBinaryEncoding(sparse, shape=(9, 9, 9))
        stripped, calculated_padding = encoding.stripped
        np.testing.assert_equal(
            stripped.sparse_indices, expected_sparse_indices)
        np.testing.assert_equal(
            calculated_padding, 2 * np.ones((3, 2), dtype=int))

    def test_empty_stripped(self):
        res = 10
        encoding = enc.DenseEncoding(np.zeros((res,) * 3, dtype=bool))
        stripped, calculated_padding = encoding.stripped
        self.assertEqual(stripped.size, 0)
        np.testing.assert_equal(
            calculated_padding, [[0, res], [0, res], [0, res]])

    def test_is_empty(self):
        res = 10
        empty = np.zeros((res,), dtype=bool)
        not_empty = np.zeros((res,), dtype=bool)
        not_empty[[1, 2, 5, 6, 7]] = True
        self.assertTrue(enc.DenseEncoding(empty).is_empty)
        self.assertFalse(enc.DenseEncoding(not_empty).is_empty)

        for cls in (
                enc.SparseEncoding,
                enc.RunLengthEncoding,
                enc.BinaryRunLengthEncoding):
            self.assertTrue(cls.from_dense(empty).is_empty)
            self.assertFalse(cls.from_dense(not_empty).is_empty)


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
xaml.py
---------

Load 3D XAMl files, an export option from Solidworks.
"""
import numpy as np

import collections

from .. import util
from .. import visual
from .. import transformations as tf


def load_XAML(file_obj, *args, **kwargs):
    """
    Load a 3D XAML file.

    Parameters
    ----------
    file_obj : file object
      Open XAML file.

    Returns
    ----------
    result : dict
      Kwargs for a Trimesh constructor.
    """
    def element_to_color(element):
        """
        Turn an XML element into a (4,) np.uint8 RGBA color
        """
        if element is None:
            return visual.DEFAULT_COLOR
        hexcolor = int(element.attrib['Color'].replace('#', ''), 16)
        opacity = float(element.attrib['Opacity'])
        rgba = [(hexcolor >> 16) & 0xFF,
                (hexcolor >> 8) & 0xFF,
                (hexcolor & 0xFF),
                opacity * 0xFF]
        rgba = np.array(rgba, dtype=np.uint8)
        return rgba

    def element_to_transform(element):
        """
        Turn an XML element into a (4,4) np.float64
        transformation matrix.
        """
        try:
            matrix = next(element.iter(
                tag=ns + 'MatrixTransform3D')).attrib['Matrix']
            matrix = np.array(matrix.split(),
                              dtype=np.float64).reshape((4, 4)).T
            return matrix
        except StopIteration:
            # this will be raised if the MatrixTransform3D isn't in the passed
            # elements tree
            return np.eye(4)

    # read the file and parse XML
    file_data = file_obj.read()
    root = etree.XML(file_data)

    # the XML namespace
    ns = root.tag.split('}')[0] + '}'

    # the linked lists our results are going in
    vertices = []
    faces = []
    colors = []
    normals = []

    # iterate through the element tree
    # the GeometryModel3D tag contains a material and geometry
    for geometry in root.iter(tag=ns + 'GeometryModel3D'):

        # get the diffuse and specular colors specified in the material
        color_search = './/{ns}{color}Material/*/{ns}SolidColorBrush'
        diffuse = geometry.find(color_search.format(ns=ns,
                                                    color='Diffuse'))
        specular = geometry.find(color_search.format(ns=ns,
                                                     color='Specular'))

        # convert the element into a (4,) np.uint8 RGBA color
        diffuse = element_to_color(diffuse)
        specular = element_to_color(specular)

        # to get the final transform of a component we'll have to traverse
        # all the way back to the root node and save transforms we find
        current = geometry
        transforms = collections.deque()
        # when the root node is reached its parent will be None and we stop
        while current is not None:
            # element.find will only return elements that are direct children
            # of the current element as opposed to element.iter,
            # which will return any depth of child
            transform_element = current.find(ns + 'ModelVisual3D.Transform')
            if transform_element is not None:
                # we are traversing the tree backwards, so append new
                # transforms to the left of the deque
                transforms.appendleft(element_to_transform(transform_element))
            # we are going from the lowest level of the tree to the highest
            # this avoids having to traverse any branches that don't have
            # geometry
            current = current.getparent()

        if len(transforms) == 0:
            # no transforms in the tree mean an identity matrix
            transform = np.eye(4)
        elif len(transforms) == 1:
            # one transform in the tree we can just use
            transform = transforms.pop()
        else:
            # multiple transforms we apply all of them in order
            transform = util.multi_dot(transforms)

        # iterate through the contained mesh geometry elements
        for g in geometry.iter(tag=ns + 'MeshGeometry3D'):
            c_normals = np.array(g.attrib['Normals'].replace(',', ' ').split(),
                                 dtype=np.float64).reshape((-1, 3))

            c_vertices = np.array(
                g.attrib['Positions'].replace(
                    ',', ' ').split(), dtype=np.float64).reshape(
                (-1, 3))
            # bake in the transform as we're saving
            c_vertices = tf.transform_points(c_vertices, transform)

            c_faces = np.array(
                g.attrib['TriangleIndices'].replace(
                    ',', ' ').split(), dtype=np.int64).reshape(
                (-1, 3))

            # save data to a sequence
            vertices.append(c_vertices)
            faces.append(c_faces)
            colors.append(np.tile(diffuse, (len(c_faces), 1)))
            normals.append(c_normals)

    # compile the results into clean numpy arrays
    result = dict()
    result['vertices'], result['faces'] = util.append_faces(vertices,
                                                            faces)
    result['face_colors'] = np.vstack(colors)
    result['vertex_normals'] = np.vstack(normals)

    return result


try:
    from lxml import etree
    _xaml_loaders = {'xaml': load_XAML}
except BaseException as E:
    # create a dummy module which will raise the ImportError
    # or other exception only when someone tries to use networkx
    from ..exceptions import ExceptionWrapper
    _xaml_loaders = {'xaml': ExceptionWrapper(E)}



try:
    from . import generic as g
except BaseException:
    import generic as g


class SegmentsTest(g.unittest.TestCase):

    def test_param(self):
        from trimesh.path import segments

        # check 2D and 3D
        for dimension in [2, 3]:
            # a bunch of random line segments
            s = g.random((100, 2, dimension))
            # convert segment to point on line closest to origin
            # as well as a vector and two distances along vector
            param = segments.segments_to_parameters(s)
            # convert parameterized back to segments
            roundtrip = segments.parameters_to_segments(*param)
            # they should be the same after a roundtrip
            assert g.np.allclose(s, roundtrip)

            # make index 1 the first segment but offset along vector
            # IE make s[0] colinear with s[1]
            s[1] = s[0] + 10 * (s[0][0] - s[0][1])
            # calculate colinear pairs
            colinear = segments.colinear_pairs(s)

            # due to our wangling the first and second index
            # should be a colinear pair
            assert {0, 1} in [set(i) for i in colinear]

    def test_colinear(self):
        from trimesh.path import segments

        seg = g.np.array([[[0., 0., 0.],
                           [0., 1., 0.]],

                          [[0., 1., 0.],
                           [0., 0., 0.]],

                          [[0., 0., 0.],
                           [0., 2., 0.]],

                          [[0., 0.5, 0.],
                           [0., 0.75, 0.]],

                          [[0., 2.1, 0.],
                           [0., 2.2, 0.]],

                          [[0., 2.0, 0.],
                           [0., 2.3, 0.]],

                          [[0., 0., 0.],
                           [1., 1., 0.]]])

        # get the unit direction vector for the segments
        unit = g.trimesh.unitize(g.np.diff(
            seg, axis=1).reshape((-1, 3)))

        L = segments.colinear_pairs(seg[:3])
        assert len(L) == 3
        # make sure all pairs are really colinear
        dots = [g.np.dot(*row) for row in unit[L]]
        assert (g.np.isclose(dots, 1.0) |
                g.np.isclose(dots, -1)).all()

        L = segments.colinear_pairs(seg)
        dots = [g.np.dot(*row) for row in unit[L]]
        assert (g.np.isclose(dots, 1.0) |
                g.np.isclose(dots, -1)).all()

        epsilon = 1e-6
        # length should only include vectors with one
        # vertex closer than epsilon
        n = segments.colinear_pairs(seg, length=epsilon)
        dots = [g.np.dot(*row) for row in unit[L]]
        assert (g.np.isclose(dots, 1.0) |
                g.np.isclose(dots, -1)).all()

        for pair in n:
            val = seg[pair]
            close = g.np.append(
                (val[0] - val[1]).ptp(axis=1),
                (val[0] - val[1][::-1]).ptp(axis=1)).min()
            assert close < epsilon

    def test_extrude(self):
        from trimesh.path.segments import extrude
        # hand tuned segments
        manual = g.np.column_stack((
            g.np.zeros((3, 2)),
            [[0, 1], [0, -1], [1, 2]])).reshape((-1, 2, 2))

        for seg in [manual, g.random((10, 2, 2))]:
            height = 1.22
            v, f = extrude(segments=seg, height=height)
            # load extrusion as mesh
            mesh = g.trimesh.Trimesh(vertices=v, faces=f)
            # load line segments as path
            path = g.trimesh.load_path(seg)
            # compare area of mesh with source path
            assert g.np.isclose(mesh.area, path.length * height)

    def test_resample(self):
        from trimesh.path.segments import length, resample
        # create some random segments
        seg = g.random((1000, 2, 3))
        # set a maximum segment length
        maxlen = 0.1
        # one of the original segments should be longer than maxlen
        assert (length(seg, summed=False) > maxlen).any()
        # resample to be all shorter than maxlen
        res = resample(seg, maxlen=maxlen)
        # check lengths of the resampled result
        assert (length(res, summed=False) < maxlen).all()
        # make sure overall length hasn't changed
        assert g.np.isclose(length(res), length(seg))

        # now try with indexes returned
        res, index = resample(seg, maxlen=maxlen, return_index=True)
        # check lengths of the resampled result
        assert (length(res, summed=False) < maxlen).all()
        # make sure overall length hasn't changed
        assert g.np.isclose(length(res), length(seg))

    def test_svg(self):
        from trimesh.path.segments import to_svg
        # create some 2D segments
        seg = g.random((1000, 2, 2))
        # make one of the segments a duplicate
        seg[0] = seg[-1]
        # create an SVG path string
        svg = to_svg(seg, merge=False)
        # should be one move and one line per segment
        assert svg.count('M') == len(seg)
        assert svg.count('L') == len(seg)

        # try with a transform
        svg = to_svg(seg, matrix=g.np.eye(3), merge=False)
        assert svg.count('M') == len(seg)
        assert svg.count('L') == len(seg)

        # remove the duplicate segments
        svg = to_svg(seg, matrix=g.np.eye(3), merge=True)
        assert svg.count('M') < len(seg)
        assert svg.count('L') < len(seg)

        try:
            to_svg(g.random((100, 2, 3)))
        except ValueError:
            return
        raise ValueError('to_svg accepted wrong input!')


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
test_minimal.py
--------------

Test things that should work with a *minimal* trimesh install.


"""
import os

import unittest
import trimesh
import numpy as np

# the path of the current directory
_pwd = os.path.dirname(
    os.path.abspath(os.path.expanduser(__file__)))
# the absolute path for our reference models
_mwd = os.path.abspath(
    os.path.join(_pwd, '..', 'models'))


def get_mesh(file_name, **kwargs):
    """
    Load a mesh from our models directory.
    """
    return trimesh.load(os.path.join(_mwd, file_name),
                        **kwargs)


class MinimalTest(unittest.TestCase):

    def test_path_exc(self):
        # this should require *no deps*
        from trimesh.path import packing
        bounds, inserted = packing.rectangles_single(
            [[1, 1], [2, 2]],
            size=[2, 4])
        assert inserted.all()

        extents = bounds.reshape((-1, 2)).ptp(axis=0)
        assert np.allclose(extents, [2, 3])
        assert bounds.shape == (2, 2, 2)
        density = 5.0 / np.product(extents)
        assert density > .833

    def test_load(self):
        # kinds of files we should be able to
        # load even with a minimal install
        kinds = 'stl ply obj off gltf glb'.split()

        for file_name in os.listdir(_mwd):
            ext = os.path.splitext(file_name)[-1].lower()[1:]
            if ext not in kinds:
                continue

            m = get_mesh(file_name)
            if isinstance(m, trimesh.Trimesh):
                assert len(m.face_adjacency.shape) == 2
                assert len(m.vertices.shape) == 2

                # make sure hash changes
                initial = hash(m)
                m.faces += 0

                assert hash(m) == initial
                m.vertices[:, 0] += 1.0
                assert hash(m) != initial

    def test_load_path(self):
        # should be able to load a path and export it as a GLB
        # try with a Path3D
        path = trimesh.load_path(np.asarray([(0, 0, 0), (1, 0, 0), (1, 1, 0)]))
        assert isinstance(path, trimesh.path.Path3D)
        scene = trimesh.Scene(path)
        assert len(scene.geometry) == 1
        glb = scene.export(file_type='glb')
        assert len(glb) > 0

        # now create a Path2D
        path = trimesh.load_path(np.asarray([(0, 0), (1, 0), (1, 1)]))
        assert isinstance(path, trimesh.path.Path2D)

        # export to an SVG
        svg = path.export(file_type='svg')
        assert len(svg) > 0

        dxf = path.export(file_type='dxf')
        assert len(dxf) > 0

        scene = trimesh.Scene(path)
        assert len(scene.geometry) == 1
        glb = scene.export(file_type='glb')
        assert len(glb) > 0

    def test_load_wrap(self):
        # check to see if we give a useful message
        # when we *do not* have `lxml` installed
        try:
            import lxml  # noqa
            return
        except BaseException:
            pass

        # we have no 3DXML
        exc = None
        try:
            get_mesh('cycloidal.3DXML')
        except BaseException as E:
            exc = str(E)

        # should have raised
        assert exc is not None

        # error message should have been useful
        assert 'lxml' in exc


if __name__ == '__main__':
    trimesh.util.attach_to_log()
    unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class SampleTest(g.unittest.TestCase):

    def test_sample(self):
        m = g.get_mesh('featuretype.STL')

        samples = m.sample(1000)

        # check to make sure all samples are on the mesh surface
        distance = m.nearest.signed_distance(samples)
        assert g.np.abs(distance).max() < 1e-4

        even, index = g.trimesh.sample.sample_surface_even(m, 1000)
        # check to make sure all samples are on the mesh surface
        distance = m.nearest.signed_distance(even)
        assert g.np.abs(distance).max() < 1e-4

    def test_weights(self):
        m = g.trimesh.creation.box()

        # weigh all faces except first as zero
        weights = g.np.zeros(len(m.faces))
        weights[0] = 1.0

        # sample with passed weights
        points, fid = m.sample(count=100,
                               return_index=True,
                               face_weight=weights)
        # all faces should be on single face
        assert (fid == 0).all()

        # oversample box to make sure weights aren't screwing
        # up ability to get every face when weighted by area
        assert set(g.np.unique(m.sample(
            100000, return_index=True)[1])) == set(range(len(m.faces)))

    def test_color(self):
        # check to see if sampling by color works

        # sample a textured mesh
        m = g.get_mesh('fuze.obj')
        points, index, color = g.trimesh.sample.sample_surface(
            m, 100, sample_color=True)
        assert len(points) == len(color)

        # sample a color mesh
        m = g.get_mesh('machinist.XAML')
        assert m.visual.kind == 'face'
        points, index, color = g.trimesh.sample.sample_surface(
            m, 100, sample_color=True)
        assert len(points) == len(color)

    def test_sample_volume(self):
        m = g.trimesh.creation.icosphere()
        samples = g.trimesh.sample.volume_mesh(mesh=m, count=100)

        # all samples should be approximatly within the sphere
        radii = g.np.linalg.norm(samples, axis=1)
        assert (radii < 1.00000001).all()

    def sample_volume_rectangular(self):
        # check to see if our OBB volume sampling runs
        m = g.get_mesh('rabbit.obj')
        obb = m.bounding_box_oriented

        # should use a box-specific volume sampling method
        samples = obb.sample_volume(100)
        assert samples.shape == (100, 3)

    def test_deterministic_sample(self):
        m = g.get_mesh('featuretype.STL')

        # Without seed passed should return non-deterministic results
        even_first, index_first = g.trimesh.sample.sample_surface(m, 10000)
        even_last, index_last = g.trimesh.sample.sample_surface(m, 10000)
        assert not (even_first == even_last).all()
        assert not (index_first == index_last).all()

        # With seed passed should return identical results
        even_first, index_first = g.trimesh.sample.sample_surface(m, 10000, seed=10)
        even_last, index_last = g.trimesh.sample.sample_surface(m, 10000, seed=10)
        assert (even_first == even_last).all()
        assert (index_first == index_last).all()


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



import numpy as np

from shapely import ops
from shapely.geometry import Polygon

from .. import bounds
from .. import graph
from .. import geometry
from .. import grouping

from ..constants import log
from ..constants import tol_path as tol
from ..transformations import transform_points

from .simplify import fit_circle_check
from .traversal import resample_path

try:
    import networkx as nx
except BaseException as E:
    # create a dummy module which will raise the ImportError
    # or other exception only when someone tries to use networkx
    from ..exceptions import ExceptionWrapper
    nx = ExceptionWrapper(E)
try:
    from rtree import Rtree
except BaseException as E:
    # create a dummy module which will raise the ImportError
    from ..exceptions import ExceptionWrapper
    Rtree = ExceptionWrapper(E)


def enclosure_tree(polygons):
    """
    Given a list of shapely polygons with only exteriors,
    find which curves represent the exterior shell or root curve
    and which represent holes which penetrate the exterior.

    This is done with an R-tree for rough overlap detection,
    and then exact polygon queries for a final result.

    Parameters
    -----------
    polygons : (n,) shapely.geometry.Polygon
       Polygons which only have exteriors and may overlap

    Returns
    -----------
    roots : (m,) int
        Index of polygons which are root
    contains : networkx.DiGraph
       Edges indicate a polygon is
       contained by another polygon
    """
    tree = Rtree()
    # nodes are indexes in polygons
    contains = nx.DiGraph()
    for i, polygon in enumerate(polygons):
        # if a polygon is None it means creation
        # failed due to weird geometry so ignore it
        if polygon is None or len(polygon.bounds) != 4:
            continue
        # insert polygon bounds into rtree
        tree.insert(i, polygon.bounds)
        # make sure every valid polygon has a node
        contains.add_node(i)

    # loop through every polygon
    for i in contains.nodes():
        polygon = polygons[i]
        # we first query for bounding box intersections from the R-tree
        for j in tree.intersection(polygon.bounds):
            # if we are checking a polygon against itself continue
            if (i == j):
                continue
            # do a more accurate polygon in polygon test
            # for the enclosure tree information
            if polygons[i].contains(polygons[j]):
                contains.add_edge(i, j)
            elif polygons[j].contains(polygons[i]):
                contains.add_edge(j, i)

    # a root or exterior curve has an even number of parents
    # wrap in dict call to avoid networkx view
    degree = dict(contains.in_degree())
    # convert keys and values to numpy arrays
    indexes = np.array(list(degree.keys()))
    degrees = np.array(list(degree.values()))
    # roots are curves with an even inward degree (parent count)
    roots = indexes[(degrees % 2) == 0]
    # if there are multiple nested polygons split the graph
    # so the contains logic returns the individual polygons
    if len(degrees) > 0 and degrees.max() > 1:
        # collect new edges for graph
        edges = []

        # order the roots so they are sorted by degree
        roots = roots[np.argsort([degree[r] for r in roots])]
        # find edges of subgraph for each root and children
        for root in roots:
            children = indexes[degrees == degree[root] + 1]
            edges.extend(contains.subgraph(
                np.append(children, root)).edges())
        # stack edges into new directed graph
        contains = nx.from_edgelist(edges, nx.DiGraph())
        # if roots have no children add them anyway
        contains.add_nodes_from(roots)

    return roots, contains


def edges_to_polygons(edges, vertices):
    """
    Given an edge list of indices and associated vertices
    representing lines, generate a list of polygons.

    Parameters
    -----------
    edges : (n, 2) int
      Indexes of vertices which represent lines
    vertices : (m, 2) float
      Vertices in 2D space

    Returns
    ----------
    polygons : (p,) shapely.geometry.Polygon
      Polygon objects with interiors
    """

    assert isinstance(vertices, np.ndarray)

    # create closed polygon objects
    polygons = []
    # loop through a sequence of ordered traversals
    for dfs in graph.traversals(edges, mode='dfs'):
        try:
            # try to recover polygons before they are more complicated
            repaired = repair_invalid(Polygon(vertices[dfs]))
            # if it returned a multipolygon extend into a flat list
            if hasattr(repaired, 'geoms'):
                polygons.extend(repaired.geoms)
            else:
                polygons.append(repaired)
        except ValueError:
            continue

    # if there is only one polygon, just return it
    if len(polygons) == 1:
        return polygons

    # find which polygons contain which other polygons
    roots, tree = enclosure_tree(polygons)

    # generate list of polygons with proper interiors
    complete = []
    for root in roots:
        interior = list(tree[root].keys())
        shell = polygons[root].exterior.coords
        holes = [polygons[i].exterior.coords for i in interior]
        complete.append(Polygon(shell=shell,
                                holes=holes))
    return complete


def polygons_obb(polygons):
    """
    Find the OBBs for a list of shapely.geometry.Polygons
    """
    rectangles = [None] * len(polygons)
    transforms = [None] * len(polygons)
    for i, p in enumerate(polygons):
        transforms[i], rectangles[i] = polygon_obb(p)
    return np.array(transforms), np.array(rectangles)


def polygon_obb(polygon):
    """
    Find the oriented bounding box of a Shapely polygon.

    The OBB is always aligned with an edge of the convex hull of the polygon.

    Parameters
    -------------
    polygons : shapely.geometry.Polygon
      Input geometry

    Returns
    -------------
    transform : (3, 3) float
      Transformation matrix
      which will move input polygon from its original position
      to the first quadrant where the AABB is the OBB
    extents : (2,) float
      Extents of transformed polygon
    """
    if hasattr(polygon, 'exterior'):
        points = np.asanyarray(polygon.exterior.coords)
    elif isinstance(polygon, np.ndarray):
        points = polygon
    else:
        raise ValueError('polygon or points must be provided')

    transform, extents = bounds.oriented_bounds_2D(points)

    if tol.strict:
        moved = transform_points(points=points, matrix=transform)
        assert np.allclose(-extents / 2.0, moved.min(axis=0))
        assert np.allclose(extents / 2.0, moved.max(axis=0))

    return transform, extents


def transform_polygon(polygon, matrix):
    """
    Transform a polygon by a a 2D homogeneous transform.

    Parameters
    -------------
    polygon : shapely.geometry.Polygon
      2D polygon to be transformed.
    matrix  : (3, 3) float
      2D homogeneous transformation.

    Returns
    --------------
    result : shapely.geometry.Polygon
      Polygon transformed by matrix.
    """
    matrix = np.asanyarray(matrix, dtype=np.float64)

    if hasattr(polygon, 'geoms'):
        result = [transform_polygon(p, t)
                  for p, t in zip(polygon, matrix)]
        return result
    # transform the outer shell
    shell = transform_points(np.array(polygon.exterior.coords),
                             matrix)[:, :2]
    # transform the interiors
    holes = [transform_points(np.array(i.coords),
                              matrix)[:, :2]
             for i in polygon.interiors]
    # create a new polygon with the result
    result = Polygon(shell=shell, holes=holes)
    return result


def polygon_bounds(polygon, matrix=None):
    """
    Get the transformed axis aligned bounding box of a
    shapely Polygon object.

    Parameters
    ------------
    polygon : shapely.geometry.Polygon
      Polygon pre-transform
    matrix : (3, 3) float or None.
      Homogeneous transform moving polygon in space

    Returns
    ------------
    bounds : (2, 2) float
      Axis aligned bounding box of transformed polygon.
    """
    if matrix is not None:
        assert matrix.shape == (3, 3)
        points = transform_points(
            points=np.array(polygon.exterior.coords),
            matrix=matrix)
    else:
        points = np.array(polygon.exterior.coords)

    bounds = np.array([points.min(axis=0),
                       points.max(axis=0)])
    assert bounds.shape == (2, 2)
    return bounds


def plot(polygon=None, show=True, axes=None, **kwargs):
    """
    Plot a shapely polygon using matplotlib.

    Parameters
    ------------
    polygon : shapely.geometry.Polygon
      Polygon to be plotted
    show : bool
      If True will display immediately
    **kwargs
      Passed to plt.plot
    """
    import matplotlib.pyplot as plt

    def plot_single(single):
        axes.plot(*single.exterior.xy, **kwargs)
        for interior in single.interiors:
            axes.plot(*interior.xy, **kwargs)
    # make aspect ratio non-stupid
    if axes is None:
        axes = plt.axes()
    axes.set_aspect('equal', 'datalim')

    if polygon.__class__.__name__ == 'MultiPolygon':
        [plot_single(i) for i in polygon.geoms]
    elif hasattr(polygon, '__iter__'):
        [plot_single(i) for i in polygon]
    elif polygon is not None:
        plot_single(polygon)

    if show:
        plt.show()

    return axes


def resample_boundaries(polygon, resolution, clip=None):
    """
    Return a version of a polygon with boundaries resampled
    to a specified resolution.

    Parameters
    -------------
    polygon : shapely.geometry.Polygon
      Source geometry
    resolution : float
      Desired distance between points on boundary
    clip : (2,) int
      Upper and lower bounds to clip
      number of samples to avoid exploding count

    Returns
    ------------
    kwargs : dict
     Keyword args for a Polygon constructor `Polygon(**kwargs)`
    """
    def resample_boundary(boundary):
        # add a polygon.exterior or polygon.interior to
        # the deque after resampling based on our resolution
        count = boundary.length / resolution
        count = int(np.clip(count, *clip))
        return resample_path(boundary.coords, count=count)
    if clip is None:
        clip = [8, 200]
    # create a sequence of [(n,2)] points
    kwargs = {'shell': resample_boundary(polygon.exterior),
              'holes': []}
    for interior in polygon.interiors:
        kwargs['holes'].append(resample_boundary(interior))

    return kwargs


def stack_boundaries(boundaries):
    """
    Stack the boundaries of a polygon into a single
    (n, 2) list of vertices.

    Parameters
    ------------
    boundaries : dict
      With keys 'shell', 'holes'

    Returns
    ------------
    stacked : (n, 2) float
      Stacked vertices
    """
    if len(boundaries['holes']) == 0:
        return boundaries['shell']
    result = np.vstack((boundaries['shell'],
                        np.vstack(boundaries['holes'])))
    return result


def medial_axis(polygon,
                resolution=None,
                clip=None):
    """
    Given a shapely polygon, find the approximate medial axis
    using a voronoi diagram of evenly spaced points on the
    boundary of the polygon.

    Parameters
    ----------
    polygon : shapely.geometry.Polygon
      The source geometry
    resolution : float
      Distance between each sample on the polygon boundary
    clip : None, or (2,) int
      Clip sample count to min of clip[0] and max of clip[1]

    Returns
    ----------
    edges : (n, 2) int
      Vertex indices representing line segments
      on the polygon's medial axis
    vertices : (m, 2) float
      Vertex positions in space
    """
    # a circle will have a single point medial axis
    if len(polygon.interiors) == 0:
        # what is the approximate scale of the polygon
        scale = np.reshape(polygon.bounds, (2, 2)).ptp(axis=0).max()
        # a (center, radius, error) tuple
        fit = fit_circle_check(
            polygon.exterior.coords, scale=scale)
        # is this polygon in fact a circle
        if fit is not None:
            # return an edge that has the center as the midpoint
            epsilon = np.clip(
                fit['radius'] / 500, 1e-5, np.inf)
            vertices = np.array(
                [fit['center'] + [0, epsilon],
                 fit['center'] - [0, epsilon]],
                dtype=np.float64)
            # return a single edge to avoid consumers needing to special case
            edges = np.array([[0, 1]], dtype=np.int64)
            return edges, vertices

    from scipy.spatial import Voronoi
    from shapely import vectorized

    if resolution is None:
        resolution = np.reshape(
            polygon.bounds, (2, 2)).ptp(axis=0).max() / 100

    # get evenly spaced points on the polygons boundaries
    samples = resample_boundaries(polygon=polygon,
                                  resolution=resolution,
                                  clip=clip)
    # stack the boundary into a (m,2) float array
    samples = stack_boundaries(samples)
    # create the voronoi diagram on 2D points
    voronoi = Voronoi(samples)
    # which voronoi vertices are contained inside the polygon
    contains = vectorized.contains(polygon, *voronoi.vertices.T)
    # ridge vertices of -1 are outside, make sure they are False
    contains = np.append(contains, False)
    # make sure ridge vertices is numpy array
    ridge = np.asanyarray(voronoi.ridge_vertices, dtype=np.int64)
    # only take ridges where every vertex is contained
    edges = ridge[contains[ridge].all(axis=1)]

    # now we need to remove uncontained vertices
    contained = np.unique(edges)
    mask = np.zeros(len(voronoi.vertices), dtype=np.int64)
    mask[contained] = np.arange(len(contained))

    # mask voronoi vertices
    vertices = voronoi.vertices[contained]
    # re-index edges
    edges_final = mask[edges]

    if tol.strict:
        # make sure we didn't screw up indexes
        assert (vertices[edges_final] -
                voronoi.vertices[edges]).ptp() < 1e-5

    return edges_final, vertices


def polygon_hash(polygon):
    """
    Return a vector containing values representitive of
    a particular polygon.

    Parameters
    ---------
    polygon : shapely.geometry.Polygon
      Input geometry

    Returns
    ---------
    hashed: (6), float
      Representitive values representing input polygon
    """
    result = np.array(
        [len(polygon.interiors),
         polygon.convex_hull.area,
         polygon.convex_hull.length,
         polygon.area,
         polygon.length,
         polygon.exterior.length],
        dtype=np.float64)
    return result


def random_polygon(segments=8, radius=1.0):
    """
    Generate a random polygon with a maximum number of sides and approximate radius.

    Parameters
    ---------
    segments : int
      The maximum number of sides the random polygon will have
    radius : float
      The approximate radius of the polygon desired

    Returns
    ---------
    polygon : shapely.geometry.Polygon
      Geometry object with random exterior and no interiors.
    """
    angles = np.sort(np.cumsum(np.random.random(
        segments) * np.pi * 2) % (np.pi * 2))
    radii = np.random.random(segments) * radius
    points = np.column_stack(
        (np.cos(angles), np.sin(angles))) * radii.reshape((-1, 1))
    points = np.vstack((points, points[0]))
    polygon = Polygon(points).buffer(0.0)
    if hasattr(polygon, 'geoms'):
        return polygon.geoms[0]
    return polygon


def polygon_scale(polygon):
    """
    For a Polygon object return the diagonal length of the AABB.

    Parameters
    ------------
    polygon : shapely.geometry.Polygon
      Source geometry

    Returns
    ------------
    scale : float
      Length of AABB diagonal
    """
    extents = np.reshape(polygon.bounds, (2, 2)).ptp(axis=0)
    scale = (extents ** 2).sum() ** .5

    return scale


def paths_to_polygons(paths, scale=None):
    """
    Given a sequence of connected points turn them into
    valid shapely Polygon objects.

    Parameters
    -----------
    paths : (n,) sequence
      Of (m, 2) float closed paths
    scale: float
      Approximate scale of drawing for precision

    Returns
    -----------
    polys : (p,) list
      Filled with Polygon or None

    """
    polygons = [None] * len(paths)
    for i, path in enumerate(paths):
        if len(path) < 4:
            # since the first and last vertices are identical in
            # a closed loop a 4 vertex path is the minimum for
            # non-zero area
            continue
        try:
            polygons[i] = repair_invalid(Polygon(path), scale)
        except ValueError:
            # raised if a polygon is unrecoverable
            continue
        except BaseException:
            log.error('unrecoverable polygon', exc_info=True)
    polygons = np.array(polygons)
    return polygons


def sample(polygon, count, factor=1.5, max_iter=10):
    """
    Use rejection sampling to generate random points inside a
    polygon.

    Parameters
    -----------
    polygon : shapely.geometry.Polygon
      Polygon that will contain points
    count : int
      Number of points to return
    factor : float
      How many points to test per loop
    max_iter : int
      Maximum number of intersection checks is:
      > count * factor * max_iter

    Returns
    -----------
    hit : (n, 2) float
      Random points inside polygon
      where n <= count
    """
    # do batch point-in-polygon queries
    from shapely import vectorized

    # get size of bounding box
    bounds = np.reshape(polygon.bounds, (2, 2))
    extents = bounds.ptp(axis=0)

    # how many points to check per loop iteration
    per_loop = int(count * factor)

    # start with some rejection sampling
    points = bounds[0] + extents * np.random.random((per_loop, 2))
    # do the point in polygon test and append resulting hits
    mask = vectorized.contains(polygon, *points.T)
    hit = [points[mask]]
    hit_count = len(hit[0])
    # if our first non-looping check got enough samples exit
    if hit_count >= count:
        return hit[0][:count]

    # if we have to do iterations loop here slowly
    for _ in range(max_iter):
        # generate points inside polygons AABB
        points = np.random.random((per_loop, 2))
        points = (points * extents) + bounds[0]
        # do the point in polygon test and append resulting hits
        mask = vectorized.contains(polygon, *points.T)
        hit.append(points[mask])
        # keep track of how many points we've collected
        hit_count += len(hit[-1])
        # if we have enough points exit the loop
        if hit_count > count:
            break

    # stack the hits into an (n,2) array and truncate
    hit = np.vstack(hit)[:count]

    return hit


def repair_invalid(polygon, scale=None, rtol=.5):
    """
    Given a shapely.geometry.Polygon, attempt to return a
    valid version of the polygon through buffering tricks.

    Parameters
    -----------
    polygon : shapely.geometry.Polygon
      Source geometry
    rtol : float
      How close does a perimeter have to be
    scale : float or None
      For numerical precision reference

    Returns
    ----------
    repaired : shapely.geometry.Polygon
      Repaired polygon

    Raises
    ----------
    ValueError
      If polygon can't be repaired
    """
    if hasattr(polygon, 'is_valid') and polygon.is_valid:
        return polygon

    # basic repair involves buffering the polygon outwards
    # this will fix a subset of problems.
    basic = polygon.buffer(tol.zero)
    # if it returned multiple polygons check the largest
    if hasattr(basic, 'geoms'):
        basic = basic.geoms[np.argmax([i.area for i in basic.geoms])]

    # check perimeter of result against original perimeter
    if basic.is_valid and np.isclose(basic.length,
                                     polygon.length,
                                     rtol=rtol):
        return basic

    if scale is None:
        distance = 0.002 * np.reshape(
            polygon.bounds, (2, 2)).ptp(axis=0).mean()
    else:
        distance = 0.002 * scale

    # if there are no interiors, we can work with just the exterior
    # ring, which is often more reliable
    if len(polygon.interiors) == 0:
        # try buffering the exterior of the polygon
        # the interior will be offset by -tol.buffer
        rings = polygon.exterior.buffer(distance).interiors
        if len(rings) == 1:
            # reconstruct a single polygon from the interior ring
            recon = Polygon(shell=rings[0]).buffer(distance)
            # check perimeter of result against original perimeter
            if recon.is_valid and np.isclose(recon.length,
                                             polygon.length,
                                             rtol=rtol):
                return recon

        # try de-deuplicating the outside ring
        points = np.array(polygon.exterior.coords)
        # remove any segments shorter than tol.merge
        # this is a little risky as if it was discretized more
        # finely than 1-e8 it may remove detail
        unique = np.append(True, (np.diff(points, axis=0)**2).sum(
            axis=1)**.5 > 1e-8)
        # make a new polygon with result
        dedupe = Polygon(shell=points[unique])
        # check result
        if dedupe.is_valid and np.isclose(dedupe.length,
                                          polygon.length,
                                          rtol=rtol):
            return dedupe

    # buffer and unbuffer the whole polygon
    buffered = polygon.buffer(distance).buffer(-distance)
    # if it returned multiple polygons check the largest
    if hasattr(buffered, 'geoms'):
        areas = np.array([b.area for b in buffered.geoms])
        return buffered.geoms[areas.argmax()]

    # check perimeter of result against original perimeter
    if buffered.is_valid and np.isclose(buffered.length,
                                        polygon.length,
                                        rtol=rtol):
        log.debug('Recovered invalid polygon through double buffering')
        return buffered

    raise ValueError('unable to recover polygon!')


def projected(mesh,
              normal,
              origin=None,
              ignore_sign=True,
              rpad=1e-5,
              apad=None,
              tol_dot=0.01,
              max_regions=200):
    """
    Project a mesh onto a plane and then extract the polygon
    that outlines the mesh projection on that plane.

    Note that this will ignore back-faces, which is only
    relevant if the source mesh isn't watertight.

    Also padding: this generates a result by unioning the
    polygons of multiple connected regions, which requires
    the polygons be padded by a distance so that a polygon
    union produces a single coherent result. This distance
    is calculated as: `apad + (rpad * scale)`

    Parameters
    ----------
    mesh : trimesh.Trimesh
      Source geometry
    check : bool
      If True make sure is flat
    normal : (3,) float
      Normal to extract flat pattern along
    origin : None or (3,) float
      Origin of plane to project mesh onto
    ignore_sign : bool
      Allow a projection from the normal vector in
      either direction: this provides a substantial speedup
      on watertight meshes where the direction is irrelevant
      but if you have a triangle soup and want to discard
      backfaces you should set this to False.
    rpad : float
      Proportion to pad polygons by before unioning
      and then de-padding result by to avoid zero-width gaps.
    apad : float
      Absolute padding to pad polygons by before unioning
      and then de-padding result by to avoid zero-width gaps.
    tol_dot : float
      Tolerance for discarding on-edge triangles.
    max_regions : int
      Raise an exception if the mesh has more than this
      number of disconnected regions to fail quickly before
      unioning.

    Returns
    ----------
    projected : shapely.geometry.Polygon or None
      Outline of source mesh

    Raises
    ---------
    ValueError
      If max_regions is exceeded
    """
    # make sure normal is a unitized copy
    normal = np.array(normal, dtype=np.float64)
    normal /= np.linalg.norm(normal)

    # the projection of each face normal onto facet normal
    dot_face = np.dot(normal, mesh.face_normals.T)
    if ignore_sign:
        # for watertight mesh speed up projection by handling side with less faces
        # check if face lies on front or back of normal
        front = dot_face > tol_dot
        back = dot_face < -tol_dot
        # divide the mesh into front facing section and back facing parts
        # and discard the faces perpendicular to the axis.
        # since we are doing a unary_union later we can use the front *or*
        # the back so we use which ever one has fewer triangles
        # we want the largest nonzero group
        count = np.array([front.sum(), back.sum()])
        if count.min() == 0:
            # if one of the sides has zero faces we need the other
            pick = count.argmax()
        else:
            # otherwise use the normal direction with the fewest faces
            pick = count.argmin()
        # use the picked side
        side = [front, back][pick]
    else:
        # if explicitly asked to care about the sign
        # only handle the front side of normal
        side = dot_face > tol_dot

    # subset the adjacency pairs to ones which have both faces included
    # on the side we are currently looking at
    adjacency_check = side[mesh.face_adjacency].all(axis=1)
    adjacency = mesh.face_adjacency[adjacency_check]

    # a sequence of face indexes that are connected
    face_groups = graph.connected_components(
        adjacency, nodes=np.nonzero(side)[0])

    # if something is goofy we may end up with thousands of
    # regions that do nothing except hang for an hour then segfault
    if len(face_groups) > max_regions:
        raise ValueError('too many disconnected groups!')

    # reshape edges into shape length of faces for indexing
    edges = mesh.edges_sorted.reshape((-1, 6))
    # transform from the mesh frame in 3D to the XY plane
    to_2D = geometry.plane_transform(
        origin=origin, normal=normal)
    # transform mesh vertices to 2D and clip the zero Z
    vertices_2D = transform_points(
        mesh.vertices, to_2D)[:, :2]

    polygons = []
    for faces in face_groups:
        # index edges by face then shape back to individual edges
        edge = edges[faces].reshape((-1, 2))
        # edges that occur only once are on the boundary
        group = grouping.group_rows(edge, require_count=1)
        # turn each region into polygons
        polygons.extend(edges_to_polygons(
            edges=edge[group], vertices=vertices_2D))

    padding = 0.0
    if apad is not None:
        # set padding by absolute value
        padding += float(apad)
    if rpad is not None:
        # get the 2D scale as the longest side of the AABB
        scale = vertices_2D.ptp(axis=0).max()
        # apply the scale-relative padding
        padding += float(rpad) * scale

    # some types of errors will lead to a bajillion disconnected
    # regions and the union will take forever to fail
    # so exit here early
    if len(polygons) > max_regions:
        raise ValueError('too many disconnected groups!')

    # if there is only one region we don't need to run a union
    elif len(polygons) == 1:
        return polygons[0]
    elif len(polygons) == 0:
        return None
    else:
        # inflate each polygon before unioning to remove zero-size
        # gaps then deflate the result after unioning by the same amount
        # note the following provides a 25% speedup but needs
        # more testing to see if it deflates to a decent looking
        # result:
        # polygon = ops.unary_union(
        #    [p.buffer(padding,
        #              join_style=2,
        #              mitre_limit=1.5)
        #     for p in polygons]).buffer(-padding)
        polygon = ops.unary_union(
            [p.buffer(padding)
             for p in polygons]).buffer(-padding)
    return polygon


def second_moments(polygon, return_centered=False):
    """
    Calculate the second moments of area of a polygon
    from the boundary.

    Parameters
    ------------
    polygon : shapely.geometry.Polygon
      Closed polygon.
    return_centered : bool
      Get second moments for a frame with origin at the centroid
      and perform a principal axis transformation.

    Returns
    ----------
    moments : (3,) float
      The values of `[Ixx, Iyy, Ixy]`
    principal_moments : (2,) float
      Principal second moments of inertia: `[Imax, Imin]`
      Only returned if `centered`.
    alpha : float
      Angle by which the polygon needs to be rotated, so the
      principal axis align with the X and Y axis.
      Only returned if `centered`.
    transform : (3, 3) float
      Transformation matrix which rotates the polygon by alpha.
      Only returned if `centered`.
    """

    transform = np.eye(3)
    if return_centered:
        # calculate centroid and move polygon
        transform[:2, 2] = - np.array(polygon.centroid.coords)
        polygon = transform_polygon(polygon, transform)

    # start with the exterior
    coords = np.array(polygon.exterior.coords)
    # shorthand the coordinates
    x1, y1 = np.vstack((coords[-1], coords[:-1])).T
    x2, y2 = coords.T
    # do vectorized operations
    v = x1 * y2 - x2 * y1
    Ixx = np.sum(v * (y1 * y1 + y1 * y2 + y2 * y2)) / 12.0
    Iyy = np.sum(v * (x1 * x1 + x1 * x2 + x2 * x2)) / 12.0
    Ixy = np.sum(v * (x1 * y2 + 2 * x1 * y1 + 2 * x2 * y2 + x2 * y1)) / 24.0

    for interior in polygon.interiors:
        coords = np.array(interior.coords)
        # shorthand the coordinates
        x1, y1 = np.vstack((coords[-1], coords[:-1])).T
        x2, y2 = coords.T
        # do vectorized operations
        v = x1 * y2 - x2 * y1
        Ixx -= np.sum(v * (y1 * y1 + y1 * y2 + y2 * y2)) / 12.0
        Iyy -= np.sum(v * (x1 * x1 + x1 * x2 + x2 * x2)) / 12.0
        Ixy -= np.sum(v * (x1 * y2 + 2 * x1 * y1 +
                      2 * x2 * y2 + x2 * y1)) / 24.0

    moments = [Ixx, Iyy, Ixy]

    if not return_centered:
        return moments

    # get the principal moments
    root = np.sqrt(((Iyy - Ixx) / 2.0)**2 + Ixy**2)
    Imax = (Ixx + Iyy) / 2.0 + root
    Imin = (Ixx + Iyy) / 2.0 - root
    principal_moments = [Imax, Imin]

    # do the principal axis transform
    if np.isclose(Ixy, 0.0, atol=1e-12):
        alpha = 0
    elif np.isclose(Ixx, Iyy):
        # prevent division by 0
        alpha = 0.25 * np.pi
    else:
        alpha = 0.5 * np.arctan(2.0 * Ixy / (Ixx - Iyy))

    # construct transformation matrix
    cos_alpha = np.cos(alpha)
    sin_alpha = np.sin(alpha)

    transform[0, 0] = cos_alpha
    transform[1, 1] = cos_alpha
    transform[0, 1] = - sin_alpha
    transform[1, 0] = sin_alpha

    return moments, principal_moments, alpha, transform



try:
    from . import generic as g
except BaseException:
    import generic as g


class SceneTests(g.unittest.TestCase):

    def test_scene(self):
        for file_name in ['cycloidal.ply', 'sphere.ply']:
            mesh = g.get_mesh(file_name)
            if mesh.units is None:
                mesh.units = 'in'

            scene_split = g.trimesh.scene.split_scene(mesh)
            scene_split.convert_units('in')
            scene_base = g.trimesh.Scene(mesh)

            # save hash of scene before concat
            pre = [scene_split.__hash__(), scene_base.__hash__()]
            # make sure hash's give the same result twice
            assert scene_split.__hash__() == pre[0]
            assert scene_base.__hash__() == pre[1]

            # __hash__ is a long int which fails isinstance in Python 2
            assert type(scene_base.__hash__()).__name__ in ('int', 'long')

            # try out scene appending
            concat = scene_split + scene_base

            # make sure concat didn't mess with original scenes
            assert scene_split.__hash__() == pre[0]
            assert scene_base.__hash__() == pre[1]

            # make sure concatenate appended things, stuff
            assert len(concat.geometry) == (len(scene_split.geometry) +
                                            len(scene_base.geometry))

            for s in [scene_split, scene_base]:
                pre = s.__hash__()
                assert len(s.geometry) > 0
                assert s.is_valid

                flattened = s.graph.to_flattened()
                g.json.dumps(flattened)
                edgelist = s.graph.to_edgelist()
                g.json.dumps(edgelist)

                assert s.bounds.shape == (2, 3)
                assert s.centroid.shape == (3,)
                assert s.extents.shape == (3,)
                assert isinstance(s.scale, float)
                assert g.trimesh.util.is_shape(s.triangles, (-1, 3, 3))
                assert len(s.triangles) == len(s.triangles_node)

                assert s.__hash__() == pre
                assert s.__hash__() is not None

                # should be some duplicate nodes
                assert len(s.duplicate_nodes) > 0

                # should be a single scene camera
                assert isinstance(s.camera, g.trimesh.scene.cameras.Camera)
                # should be autogenerated lights
                assert len(s.lights) > 0
                # all lights should be lights
                assert all(isinstance(L, g.trimesh.scene.lighting.Light)
                           for L in s.lights)
                # all lights should be added to scene graph
                assert all(L.name in s.graph for L in s.lights)

                # should have put a transform in scene graph
                assert s.camera.name in s.graph

                r = s.dump()

                gltf = s.export(file_type='gltf')
                assert isinstance(gltf, dict)
                assert len(gltf) > 0
                assert len(gltf['model.gltf']) > 0

                glb = s.export(file_type='glb')
                assert len(glb) > 0
                assert isinstance(glb, bytes)

                for export_format in ['dict', 'dict64']:
                    # try exporting the scene as a dict
                    # then make sure json can serialize it
                    e = g.json.dumps(s.export(file_type=export_format))
                    # reconstitute the dict into a scene
                    r = g.trimesh.load(g.json.loads(e))

                    # make sure the extents are similar before and after
                    assert g.np.allclose(
                        g.np.product(s.extents),
                        g.np.product(r.extents))

                # move the scene to origin
                s.rezero()
                # if our cache dump was bad this will fail
                assert g.np.allclose(s.centroid, 0, atol=1e-5)

                # make sure explode doesn't crash
                s.explode()

    def test_scaling(self):
        # Test the scaling of scenes including unit conversion.

        scene = g.get_mesh('cycloidal.3DXML')

        hash_val = scene.__hash__()
        extents = scene.bounding_box_oriented.primitive.extents.copy()

        # TODO: have OBB return sorted extents
        # and adjust the transform to be correct

        factor = 10.0
        scaled = scene.scaled(factor)

        # the oriented bounding box should scale exactly
        # with the scaling factor
        assert g.np.allclose(
            scaled.bounding_box_oriented.primitive.extents /
            extents,
            factor)

        # check bounding primitives
        assert scene.bounding_box.volume > 0.0
        assert scene.bounding_primitive.volume > 0.0

        # we shouldn't have modified the original scene
        assert scene.__hash__() == hash_val
        assert scaled.__hash__() != hash_val

        # 3DXML comes in as mm
        assert all(m.units == 'mm'
                   for m in scene.geometry.values())
        assert scene.units == 'mm'

        converted = scene.convert_units('in')

        assert g.np.allclose(
            converted.bounding_box_oriented.primitive.extents / extents,
            1.0 / 25.4,
            atol=1e-3)

        # shouldn't have changed the original extents
        assert g.np.allclose(
            extents,
            scene.bounding_box_oriented.primitive.extents)

        # original shouldn't have changed
        assert converted.units == 'in'
        assert all(m.units == 'in' for m in converted.geometry.values())

        assert scene.units == 'mm'

        # we shouldn't have modified the original scene
        assert scene.__hash__() == hash_val
        assert converted.__hash__() != hash_val

    def test_scaling_3D(self):
        scene = g.get_mesh('cycloidal.3DXML')
        extents = scene.extents.copy()

        factor = [0.2, 1.3, 3.3]
        scaled = scene.scaled(factor)

        assert g.np.allclose(
            scaled.extents /
            extents,
            factor)

        factor = [3.0, 3.0, 3.0]
        scaled = scene.scaled(factor)

        assert g.np.allclose(
            scaled.extents /
            extents,
            factor)

    def test_scaling_3D_mixed(self):
        # same as test_scaling_3D but input scene contains 2D and 3D geometry
        scene = g.get_mesh('scenes.zip', mixed=True)
        extents = scene.extents.copy()

        factor = [0.2, 1.3, 3.3]
        scaled = scene.scaled(factor)

        assert g.np.allclose(
            scaled.extents /
            extents,
            factor)

        factor = [3.0, 3.0, 3.0]
        scaled = scene.scaled(factor)

        assert g.np.allclose(
            scaled.extents /
            extents,
            factor)

    def test_add_geometry(self):
        # list-typed geometry should create multiple nodes,
        # e.g., below code is equivalent to
        #     scene.add_geometry(geometry[0], node_name='voxels')
        #     scene.add_geometry(geometry[1], parent_node_name='voxels')
        scene = g.trimesh.Scene()
        geometry = [g.trimesh.creation.box(), g.trimesh.creation.box()]
        scene.add_geometry(geometry)
        assert len(scene.graph.nodes_geometry) == 2

    def test_add_concat(self):
        # create a scene with just a box in it
        a = g.trimesh.creation.box().scene()
        # do the same but move the box first
        b = g.trimesh.creation.box().apply_translation([2, 2, 2]).scene()
        # add the second scene to the first
        a.add_geometry(b)
        assert len(b.geometry) == 1
        assert len(a.geometry) == 2
        assert len(a.graph.nodes_geometry) == 2

    def test_delete(self):
        # check to make sure our geometry delete cleans up
        a = g.trimesh.creation.icosphere()
        b = g.trimesh.creation.icosphere().apply_translation([2, 0, 0])
        s = g.trimesh.Scene({'a': a, 'b': b})

        assert len(s.geometry) == 2
        assert len(s.graph.nodes_geometry) == 2
        # make sure every node has a transform
        [s.graph[n] for n in s.graph.nodes]

        # delete a geometry
        s.delete_geometry('a')
        assert len(s.geometry) == 1
        assert len(s.graph.nodes_geometry) == 1
        # if we screwed up the delete this will crash
        [s.graph[n] for n in s.graph.nodes]

    def test_dupe(self):
        m = g.get_mesh('tube.obj',
                       merge_norm=True,
                       merge_tex=True)

        assert m.body_count == 1

        s = g.trimesh.scene.split_scene(m)
        assert len(s.graph.nodes) == 2
        assert len(s.graph.nodes_geometry) == 1
        assert len(s.duplicate_nodes) == 1
        assert len(s.duplicate_nodes[0]) == 1

        c = s.copy()
        assert len(c.graph.nodes) == 2
        assert len(c.graph.nodes_geometry) == 1
        assert len(c.duplicate_nodes) == 1
        assert len(c.duplicate_nodes[0]) == 1

        u = s.convert_units('in', guess=True)
        assert len(u.graph.nodes_geometry) == 1
        assert len(u.duplicate_nodes) == 1
        assert len(u.duplicate_nodes[0]) == 1

    def test_dedupe(self):
        # create a scene with two identical meshes
        a = g.trimesh.creation.box()
        b = g.trimesh.creation.box().apply_translation([2, 2, 2])
        s = g.trimesh.Scene([a, b])

        # should have 2 geometries
        assert len(s.geometry) == 2
        assert len(s.graph.nodes_geometry) == 2

        # get a de-duplicated scene
        d = s.deduplicated()
        # should not have mutated original
        assert len(s.geometry) == 2
        assert len(s.graph.nodes_geometry) == 2
        # should only have one geometry
        assert len(d.geometry) == 1
        assert len(d.graph.nodes_geometry) == 1

    def test_3DXML(self):
        s = g.get_mesh('rod.3DXML')
        assert len(s.geometry) == 3
        assert len(s.graph.nodes_geometry) == 29

        dupe = s.duplicate_nodes
        assert len(dupe) == 3
        assert sum(len(i) for i in dupe) == 29

        # test cache dumping and survivability of bad
        # non-existent geometry specified in node_geometry
        s.graph.update(dupe[0][0], geometry='GARBAGE')
        # make sure geometry was updated
        assert s.graph[dupe[0][0]][1] == 'GARBAGE'
        # get the regenerated duplicates
        dupe = s.duplicate_nodes
        assert len(dupe) == 3
        # should have been cleanly dropped
        assert sum(len(i) for i in dupe) == 28

    def test_tri(self):
        scene = g.get_mesh('cycloidal.3DXML')

        # scene should have triangles
        assert g.trimesh.util.is_shape(scene.triangles, (-1, 3, 3))
        assert len(scene.triangles_node) == len(scene.triangles)

        # node name of inserted 2D geometry
        node = scene.add_geometry(g.get_mesh('2D/wrench.dxf'))
        # should be in the graph
        assert node in scene.graph.nodes
        # should have geometry defined
        assert node in scene.graph.nodes_geometry

        # 2D geometry has no triangles
        assert node not in scene.triangles_node
        # every geometry node except the one 2D thing
        # we inserted should be in triangles_node
        assert len(set(scene.triangles_node)) == len(
            scene.graph.nodes_geometry) - 1

    def test_empty(self):
        m = g.get_mesh('bunny.ply')

        # not watertight so will result in empty scene
        s = g.trimesh.scene.split_scene(m)
        assert len(s.geometry) == 0

        s = s.convert_units('inches')
        n = s.duplicate_nodes
        assert len(n) == 0

    def test_zipped(self):
        # Make sure a zip file with multiple file types
        # is returned as a single scene.

        # allow mixed 2D and 3D geometry
        m = g.get_mesh('scenes.zip', mixed=True)
        assert len(m.geometry) >= 6

        assert len(m.graph.nodes_geometry) >= 10
        assert any(isinstance(i, g.trimesh.path.Path2D)
                   for i in m.geometry.values())
        assert any(isinstance(i, g.trimesh.Trimesh)
                   for i in m.geometry.values())

        m = g.get_mesh('scenes.zip', mixed=False)
        assert len(m.geometry) < 6

    def test_doubling(self):
        s = g.get_mesh('cycloidal.3DXML')

        # make sure we parked our car where we thought
        assert len(s.geometry) == 13

        # concatenate a scene with itself
        r = s + s

        # new scene should have twice as much geometry
        assert len(r.geometry) == (2 * len(s.geometry))

        assert g.np.allclose(s.extents,
                             r.extents)

        # duplicate node groups should be twice as long
        set_ori = set([len(i) * 2 for i in s.duplicate_nodes])
        set_dbl = set([len(i) for i in r.duplicate_nodes])

        assert set_ori == set_dbl

    def test_empty_scene(self):
        # test an empty scene
        empty = g.trimesh.Trimesh([], [])
        assert empty.bounds is None
        assert empty.extents is None
        assert g.np.isclose(empty.scale, 1.0)

        # create a sphere
        sphere = g.trimesh.creation.icosphere()

        # empty scene should have None for bounds
        scene = empty.scene()
        assert scene.bounds is None

        # add a sphere to the empty scene
        scene.add_geometry(sphere)
        # bounds should now be populated
        assert scene.bounds.shape == (2, 3)
        assert g.np.allclose(scene.bounds, sphere.bounds)

    def test_transform(self):
        # check transforming scenes
        scene = g.trimesh.creation.box()
        assert g.np.allclose(scene.bounds, [[-.5, -.5, -.5], [.5, .5, .5]])

        scene.apply_translation([1, 0, 1])
        assert g.np.allclose(scene.bounds, [[.5, -.5, .5], [1.5, .5, 1.5]])

    def test_material_group(self):
        # check scene is correctly grouped by materials
        s = g.get_mesh('box.obj', group_material=True)
        assert set(s.geometry.keys()) == {'Material', 'SecondMaterial'}
        assert len(s.geometry['Material'].faces) == 8
        assert len(s.geometry['SecondMaterial'].faces) == 4

        # make sure our flag does something
        s = g.get_mesh('box.obj', group_material=False)
        assert set(s.geometry.keys()) != {'Material', 'SecondMaterial'}

    def test_strip(self):
        m = g.get_mesh('cycloidal.3DXML')
        assert any(g.visual.kind is not None
                   for g in m.geometry.values())
        m.strip_visuals()
        assert all(g.visual.kind is None
                   for g in m.geometry.values())

    def test_export_concat(self):
        # Scenes exported in mesh formats should be
        # concatenating the meshes somewhere.
        original = g.trimesh.creation.icosphere(
            radius=0.123312)
        original_hash = original.identifier_hash

        scene = g.trimesh.Scene()
        scene.add_geometry(original)

        with g.TemporaryDirectory() as d:
            for ext in ['stl', 'ply']:
                file_name = g.os.path.join(d, 'mesh.' + ext)
                scene.export(file_name)
                loaded = g.trimesh.load(file_name)
                assert g.np.isclose(loaded.volume,
                                    original.volume)
        # nothing should have changed
        assert original.identifier_hash == original_hash

    def test_exact_bounds(self):
        m = g.get_mesh('cycloidal.3DXML')
        assert isinstance(m, g.trimesh.Scene)

        dump = m.dump(concatenate=True)
        assert isinstance(dump, g.trimesh.Trimesh)

        # scene bounds should exactly match mesh bounds
        assert g.np.allclose(m.bounds, dump.bounds)

    def test_concatenate_mixed(self):
        scene = g.trimesh.Scene([g.trimesh.creation.icosphere(),
                                 g.trimesh.path.creation.rectangle([[0, 0], [1, 1]])])

        dump = scene.dump(concatenate=True)
        assert isinstance(dump, g.trimesh.Trimesh)

    def test_append_scenes(self):
        scene_0 = g.trimesh.Scene(base_frame='not_world')
        scene_1 = g.trimesh.Scene(base_frame='not_world')

        scene_sum = g.trimesh.scene.scene.append_scenes(
            (scene_0, scene_1), common=['not_world'], base_frame='not_world')

        assert scene_sum.graph.base_frame == 'not_world'

    def test_scene_concat(self):
        # check that primitives get upgraded to meshes
        a = g.trimesh.Scene([g.trimesh.primitives.Sphere(center=[5,5,5]),
                           g.trimesh.primitives.Box()])
        c = a.dump(concatenate=True)
        assert isinstance(c, g.trimesh.Trimesh)
        assert g.np.allclose(c.bounds, a.bounds)

        c = a.dump(concatenate=False)
        assert len(c) == len(a.geometry)

        # scene 2D
        scene_2D = g.trimesh.Scene(g.get_mesh('2D/250_cycloidal.DXF').split())
        concat = scene_2D.dump(concatenate=True)
        assert isinstance(concat, g.trimesh.path.Path2D)


        dump = scene_2D.dump(concatenate=False)
        assert len(dump) == len(scene_2D.geometry)
        assert all(isinstance(i, g.trimesh.path.Path2D) for i in dump)

        # all Path3D objects
        scene_3D = g.trimesh.Scene(
            [i.to_3D() for i in
             g.get_mesh('2D/250_cycloidal.DXF').split()])

        dump= scene_3D.dump(concatenate=False)
        assert len(dump) >= 5
        assert all(isinstance(i, g.trimesh.path.Path3D) for i in dump)

        concat = scene_3D.dump(concatenate=True)
        assert isinstance(concat, g.trimesh.path.Path3D)

        mixed = list(scene_2D.geometry.values())
        mixed.extend(scene_3D.geometry.values())
        scene_mixed = g.trimesh.Scene(mixed)

        dump = scene_mixed.dump(concatenate=False)
        assert len(dump) == len(mixed)

        concat = scene_mixed.dump(concatenate=True)
        assert isinstance(concat, g.trimesh.path.Path3D)


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class ResolverTest(g.unittest.TestCase):

    def test_filepath_namespace(self):
        # check the namespaced method
        models = g.dir_models
        subdir = '2D'

        # create a resolver for the models diretory
        resolver = g.trimesh.resolvers.FilePathResolver(models)

        # should be able to get an asset
        assert len(resolver.get('rabbit.obj')) > 0

        # check a few file path keys
        check = set(['ballA.off', 'featuretype.STL'])
        assert set(resolver.keys()).issuperset(check)

        # try a namespaced resolver
        ns = resolver.namespaced(subdir)
        assert not set(ns.keys()).issuperset(check)
        assert set(ns.keys()).issuperset(['tray-easy1.dxf',
                                          'single_arc.dxf'])

    def test_web_namespace(self):
        base = 'https://example.com'
        name = 'stuff'
        target = 'hi.gltf'

        # check with a trailing slash
        a = g.trimesh.resolvers.WebResolver(base + '/')
        b = g.trimesh.resolvers.WebResolver(base + '//')
        c = g.trimesh.resolvers.WebResolver(base)
        d = a.namespaced(name)

        # base URL's should always be the same with one trailing slash
        assert a.base_url == b.base_url
        assert b.base_url == c.base_url
        assert c.base_url == base + '/'
        # check namespaced
        assert d.base_url == base + '/' + name + '/'

        # should have correct slashes
        truth = '/'.join([base, name, target])

        assert a.base_url + name + '/' + target == truth
        assert d.base_url + target == truth

    def test_items(self):
        # check __getitem__ and __setitem__
        archive = {}
        resolver = g.trimesh.resolvers.ZipResolver(archive)
        assert len(set(resolver.keys())) == 0
        resolver['hi'] = b'what'
        # should have one item
        assert set(resolver.keys()) == set(['hi'])
        # should have the right value
        assert resolver['hi'] == b'what'
        # original archive should have been modified
        assert set(archive.keys()) == set(['hi'])

        # add a subdirectory key
        resolver['stuff/nah'] = b'sup'
        assert set(archive.keys()) == set(['hi', 'stuff/nah'])
        assert set(resolver.keys()) == set(['hi', 'stuff/nah'])

        # try namespacing
        ns = resolver.namespaced('stuff')
        assert ns['nah'] == b'sup'
        g.log.debug(ns.keys())
        assert set(ns.keys()) == set(['nah'])


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
remesh.py
-------------

Deal with re- triangulation of existing meshes.
"""
import numpy as np

from . import util
from . import grouping
from . import graph

from .geometry import faces_to_edges
from .constants import tol


def subdivide(vertices,
              faces,
              face_index=None,
              vertex_attributes=None,
              return_index=False):
    """
    Subdivide a mesh into smaller triangles.

    Note that if `face_index` is passed, only those
    faces will be subdivided and their neighbors won't
    be modified making the mesh no longer "watertight."

    Parameters
    ------------
    vertices : (n, 3) float
      Vertices in space
    faces : (m, 3) int
      Indexes of vertices which make up triangular faces
    face_index : faces to subdivide.
      if None: all faces of mesh will be subdivided
      if (n,) int array of indices: only specified faces
    vertex_attributes : dict
      Contains (n, d) attribute data
    return_index : bool
      If True, return index of original face for new faces

    Returns
    ----------
    new_vertices : (q, 3) float
      Vertices in space
    new_faces : (p, 3) int
      Remeshed faces
    index_dict : dict
      Only returned if `return_index`, {index of
      original face : index of new faces}.
    """
    if face_index is None:
        face_mask = np.ones(len(faces), dtype=bool)
    else:
        face_mask = np.zeros(len(faces), dtype=bool)
        face_mask[face_index] = True

    # the (c, 3) int array of vertex indices
    faces_subset = faces[face_mask]

    # find the unique edges of our faces subset
    edges = np.sort(faces_to_edges(faces_subset), axis=1)
    unique, inverse = grouping.unique_rows(edges)
    # then only produce one midpoint per unique edge
    mid = vertices[edges[unique]].mean(axis=1)
    mid_idx = inverse.reshape((-1, 3)) + len(vertices)

    # the new faces_subset with correct winding
    f = np.column_stack([faces_subset[:, 0],
                         mid_idx[:, 0],
                         mid_idx[:, 2],
                         mid_idx[:, 0],
                         faces_subset[:, 1],
                         mid_idx[:, 1],
                         mid_idx[:, 2],
                         mid_idx[:, 1],
                         faces_subset[:, 2],
                         mid_idx[:, 0],
                         mid_idx[:, 1],
                         mid_idx[:, 2]]).reshape((-1, 3))

    # add the 3 new faces_subset per old face all on the end
    # by putting all the new faces after all the old faces
    # it makes it easier to understand the indexes
    new_faces = np.vstack((faces[~face_mask], f))
    # stack the new midpoint vertices on the end
    new_vertices = np.vstack((vertices, mid))

    if vertex_attributes is not None:
        new_attributes = {}
        for key, values in vertex_attributes.items():
            attr_tris = values[faces_subset]
            attr_mid = np.vstack([
                attr_tris[:, g, :].mean(axis=1)
                for g in [[0, 1],
                          [1, 2],
                          [2, 0]]])
            attr_mid = attr_mid[unique]
            new_attributes[key] = np.vstack((
                values, attr_mid))
        return new_vertices, new_faces, new_attributes

    if return_index:
        # turn the mask back into integer indexes
        nonzero = np.nonzero(face_mask)[0]
        # new faces start past the original faces
        # but we've removed all the faces in face_mask
        start = len(faces) - len(nonzero)
        # indexes are just offset from start
        stack = np.arange(
            start, start + len(f) * 4).reshape((-1, 4))
        # reformat into a slightly silly dict for some reason
        index_dict = {k: v for k, v in zip(nonzero, stack)}

        return new_vertices, new_faces, index_dict

    return new_vertices, new_faces


def subdivide_to_size(vertices,
                      faces,
                      max_edge,
                      max_iter=10,
                      return_index=False):
    """
    Subdivide a mesh until every edge is shorter than a
    specified length.

    Will return a triangle soup, not a nicely structured mesh.

    Parameters
    ------------
    vertices : (n, 3) float
      Vertices in space
    faces : (m, 3) int
      Indices of vertices which make up triangles
    max_edge : float
      Maximum length of any edge in the result
    max_iter : int
      The maximum number of times to run subdivision
    return_index : bool
      If True, return index of original face for new faces

    Returns
    ------------
    vertices : (j, 3) float
      Vertices in space
    faces : (q, 3) int
      Indices of vertices
    index : (q, 3) int
      Only returned if `return_index`, index of
      original face for each new face.
    """
    # store completed
    done_face = []
    done_vert = []
    done_idx = []

    # copy inputs and make sure dtype is correct
    current_faces = np.array(
        faces, dtype=np.int64, copy=True)
    current_vertices = np.array(
        vertices, dtype=np.float64, copy=True)

    # store a map to the original face index
    current_index = np.arange(len(faces))

    # loop through iteration cap
    for i in range(max_iter + 1):
        # compute the length of every triangle edge
        edge_length = (np.diff(
            current_vertices[current_faces[:, [0, 1, 2, 0]], :3],
            axis=1) ** 2).sum(axis=2) ** 0.5
        # check edge length against maximum
        too_long = (edge_length > max_edge).any(axis=1)
        # faces that are OK
        face_ok = ~too_long

        # clean up the faces a little bit so we don't
        # store a ton of unused vertices
        unique, inverse = grouping.unique_bincount(
            current_faces[face_ok].flatten(),
            return_inverse=True)

        # store vertices and faces meeting criteria
        done_vert.append(current_vertices[unique])
        done_face.append(inverse.reshape((-1, 3)))

        if return_index:
            done_idx.append(current_index[face_ok])
            current_index = np.tile(current_index[too_long],
                                    (4, 1)).T.ravel()

        # met our goals so exit
        if not too_long.any():
            break

        # check max_iter before subdividing again
        if i >= max_iter:
            raise ValueError('max_iter exceeded!')

        # run subdivision again
        (current_vertices,
         current_faces) = subdivide(current_vertices,
                                    current_faces[too_long])

    # stack sequence into nice (n, 3) arrays
    final_vertices, final_faces = util.append_faces(
        done_vert, done_face)

    if return_index:
        final_index = np.concatenate(done_idx)
        assert len(final_index) == len(final_faces)
        return final_vertices, final_faces, final_index

    return final_vertices, final_faces


def subdivide_loop(vertices,
                   faces,
                   iterations=None):
    """
    Subdivide a mesh by dividing each triangle into four triangles
    and approximating their smoothed surface (loop subdivision).
    This function is an array-based implementation of loop subdivision,
    which avoids slow for loop and enables faster calculation.

    Overall process:
    1. Calculate odd vertices.
      Assign a new odd vertex on each edge and
      calculate the value for the boundary case and the interior case.
      The value is calculated as follows.
          v2
        / f0 \\        0
      v0--e--v1      /   \\
        \\f1 /     v0--e--v1
          v3
      - interior case : 3:1 ratio of mean(v0,v1) and mean(v2,v3)
      - boundary case : mean(v0,v1)
    2. Calculate even vertices.
      The new even vertices are calculated with the existing
      vertices and their adjacent vertices.
        1---2
       / \\/ \\      0---1
      0---v---3     / \\/ \\
       \\ /\\/    b0---v---b1
        k...4
      - interior case : (1-kB):B ratio of v and k adjacencies
      - boundary case : 3:1 ratio of v and mean(b0,b1)
    3. Compose new faces with new vertices.

    Parameters
    ------------
    vertices : (n, 3) float
      Vertices in space
    faces : (m, 3) int
      Indices of vertices which make up triangles

    Returns
    ------------
    vertices : (j, 3) float
      Vertices in space
    faces : (q, 3) int
      Indices of vertices
    iterations : int
          Number of iterations to run subdivision
    """
    try:
        from itertools import zip_longest
    except BaseException:
        # python2
        from itertools import izip_longest as zip_longest

    if iterations is None:
        iterations = 1

    def _subdivide(vertices, faces):
        # find the unique edges of our faces
        edges, edges_face = faces_to_edges(
            faces, return_index=True)
        edges.sort(axis=1)
        unique, inverse = grouping.unique_rows(edges)

        # set interior edges if there are two edges and boundary if there is
        # one.
        edge_inter = np.sort(
            grouping.group_rows(
                edges,
                require_count=2),
            axis=1)
        edge_bound = grouping.group_rows(edges, require_count=1)
        # make sure that one edge is shared by only one or two faces.
        if not len(edge_inter) * 2 + len(edge_bound) == len(edges):
            # we have multiple bodies it's a party!
            # edges shared by 2 faces are "connected"
            # so this connected components operation is
            # essentially identical to `face_adjacency`
            faces_group = graph.connected_components(
                edges_face[edge_inter])

            if len(faces_group) == 1:
                raise ValueError('Some edges are shared by more than 2 faces')

            # collect a subdivided copy of each body
            seq_verts = []
            seq_faces = []
            # keep track of vertex count as we go so
            # we can do a single vstack at the end
            count = 0
            # loop through original face indexes
            for f in faces_group:
                # a lot of the complexity in this operation
                # is computing vertex neighbors so we only
                # want to pass forward the referenced vertices
                # for this particular group of connected faces
                unique, inverse = grouping.unique_bincount(
                    faces[f].reshape(-1), return_inverse=True)

                # subdivide this subset of faces
                cur_verts, cur_faces = _subdivide(
                    vertices=vertices[unique],
                    faces=inverse.reshape((-1, 3)))

                # increment the face references to match
                # the vertices when we stack them later
                cur_faces += count
                # increment the total vertex count
                count += len(cur_verts)
                # append to the sequence
                seq_verts.append(cur_verts)
                seq_faces.append(cur_faces)

            # return results as clean (n, 3) arrays
            return np.vstack(seq_verts), np.vstack(seq_faces)

        # set interior, boundary mask for unique edges
        edge_bound_mask = np.zeros(len(edges), dtype=bool)
        edge_bound_mask[edge_bound] = True
        edge_bound_mask = edge_bound_mask[unique]
        edge_inter_mask = ~edge_bound_mask

        # find the opposite face for each edge
        edge_pair = np.zeros(len(edges)).astype(int)
        edge_pair[edge_inter[:, 0]] = edge_inter[:, 1]
        edge_pair[edge_inter[:, 1]] = edge_inter[:, 0]
        opposite_face1 = edges_face[unique]
        opposite_face2 = edges_face[edge_pair[unique]]

        # set odd vertices to the middle of each edge (default as boundary
        # case).
        odd = vertices[edges[unique]].mean(axis=1)
        # modify the odd vertices for the interior case
        e = edges[unique[edge_inter_mask]]
        e_v0 = vertices[e][:, 0]
        e_v1 = vertices[e][:, 1]
        e_f0 = faces[opposite_face1[edge_inter_mask]]
        e_f1 = faces[opposite_face2[edge_inter_mask]]
        e_v2_idx = e_f0[~(e_f0[:, :, None] == e[:, None, :]).any(-1)]
        e_v3_idx = e_f1[~(e_f1[:, :, None] == e[:, None, :]).any(-1)]
        e_v2 = vertices[e_v2_idx]
        e_v3 = vertices[e_v3_idx]

        # simplified from:
        # # 3 / 8 * (e_v0 + e_v1) + 1 / 8 * (e_v2 + e_v3)
        odd[edge_inter_mask] = 0.375 * e_v0 + \
            0.375 * e_v1 + e_v2 / 8.0 + e_v3 / 8.0

        # find vertex neighbors of each vertex
        neighbors = graph.neighbors(
            edges=edges[unique],
            max_index=len(vertices))
        # convert list type of array into a fixed-shaped numpy array (set -1 to
        # empties)
        neighbors = np.array(list(zip_longest(*neighbors, fillvalue=-1))).T
        # if the neighbor has -1 index, its point is (0, 0, 0), so that
        # it is not included in the summation of neighbors when calculating the
        # even
        vertices_ = np.vstack([vertices, [0.0, 0.0, 0.0]])
        # number of neighbors
        k = (neighbors + 1).astype(bool).sum(axis=1)

        # calculate even vertices for the interior case
        even = np.zeros_like(vertices)

        # beta = 1 / k * (5 / 8 - (3 / 8 + 1 / 4 * np.cos(2 * np.pi / k)) ** 2)
        # simplified with sympy.parse_expr('...').simplify()
        beta = (40.0 - (2.0 * np.cos(2 * np.pi / k) + 3)**2) / (64 * k)
        even = beta[:, None] * vertices_[neighbors].sum(1) \
            + (1 - k[:, None] * beta[:, None]) * vertices

        # calculate even vertices for the boundary case
        if edge_bound_mask.any():
            # boundary vertices from boundary edges
            vrt_bound_mask = np.zeros(len(vertices), dtype=bool)
            vrt_bound_mask[np.unique(edges[unique][~edge_inter_mask])] = True
            # one boundary vertex has two neighbor boundary vertices (set
            # others as -1)
            boundary_neighbors = neighbors[vrt_bound_mask]
            boundary_neighbors[~vrt_bound_mask[neighbors[vrt_bound_mask]]] = -1

            even[vrt_bound_mask] = (vertices_[boundary_neighbors].sum(axis=1) / 8.0 +
                                    (3.0 / 4.0) * vertices[vrt_bound_mask])

        # the new faces with odd vertices
        odd_idx = inverse.reshape((-1, 3)) + len(vertices)
        new_faces = np.column_stack([
            faces[:, 0],
            odd_idx[:, 0],
            odd_idx[:, 2],
            odd_idx[:, 0],
            faces[:, 1],
            odd_idx[:, 1],
            odd_idx[:, 2],
            odd_idx[:, 1],
            faces[:, 2],
            odd_idx[:, 0],
            odd_idx[:, 1],
            odd_idx[:, 2]]).reshape((-1, 3))

        # stack the new even vertices and odd vertices
        new_vertices = np.vstack((even, odd))

        return new_vertices, new_faces

    for _ in range(iterations):
        vertices, faces = _subdivide(vertices, faces)

    if tol.strict or True:
        assert np.isfinite(vertices).all()
        assert np.isfinite(faces).all()
        # should raise if faces are malformed
        assert np.isfinite(vertices[faces]).all()

        # none of the faces returned should be degenerate
        # i.e. every face should have 3 unique vertices
        assert (faces[:, 1:] != faces[:, :1]).all()

    return vertices, faces



"""
trimesh.path
-------------

Handle 2D and 3D vector paths such as those contained in an
SVG or DXF file.
"""
try:
    from .path import Path2D, Path3D
except BaseException as E:
    from .. import exceptions
    Path2D = exceptions.ExceptionWrapper(E)
    Path3D = exceptions.ExceptionWrapper(E)

# explicitly add objects to all as per pep8
__all__ = ['Path2D', 'Path3D']
